from contextlib import contextmanager
from typing import Collection, List, Set, Tuple, Union

from compmake.exceptions import CompmakeDBError
from zuper_commons.types import check_isinstance, raise_desc, raise_wrapped
from .dependencies import collect_dependencies
from .queries import jobs_defined
from .storage import get_job_userobject
from ..exceptions import CompmakeBug
from ..structures import Cache, CMJobID, Job
from ..utils import memoized_reset
from compmake import CompmakeConstants, logger

__all__ = [
    "CacheQueryDB",
    "direct_uptodate_deps_inverse_closure",
    "up_to_date",
    "direct_uptodate_deps_inverse",
]


# @contract(returns="tuple(bool, unicode)")
def up_to_date(job_id: CMJobID, db) -> Tuple[bool, str]:
    """

    Check that the job is up to date.
    We are up to date if:
    *) we are in the up_to_date_cache
       (nothing uptodate can become not uptodate so this is generally safe)
    OR
    1) we have a cache AND the timestamp is not 0 (force remake) or -1 (temp)
    2) the children are up to date AND

    3a) Original case:

        the children timestamp is older than this timestamp

    3b) New strategy

        the hash of the cache is the same as the hash of the arguments

    Returns a pair:

        boolean, explanation

    """
    cq = CacheQueryDB(db)
    res, reason, _ = cq.up_to_date(job_id)
    return res, reason


def direct_uptodate_deps(job_id: CMJobID, db) -> Set[CMJobID]:
    """ Returns all direct 'dependencies' of this job:
        the jobs that are children (arguemnts)
        plus the job that defined it (if not root).
    """
    from compmake.jobs.queries import direct_children

    dependencies = direct_children(job_id, db)

    # plus jobs that defined it
    from compmake.jobs.storage import get_job

    defined_by = get_job(job_id, db).defined_by
    last = defined_by[-1]
    if last != "root":
        dependencies.add(last)
    return dependencies


def direct_uptodate_deps_inverse(job_id: CMJobID, db) -> Set[CMJobID]:
    """ Returns all jobs that have this as
        a direct 'dependency'
        the jobs that are direct parents
        plus the jobs that were defined by it.

        Assumes that the job is DONE.
    """
    from compmake.jobs.queries import direct_parents

    dep_inv = direct_parents(job_id, db)
    from compmake.jobs.storage import get_job_cache

    # Not sure if need to be here --- added when doing graph-animation for jobs in progress
    if get_job_cache(job_id, db).state == Cache.DONE:
        dep_inv.update(jobs_defined(job_id, db))
    return dep_inv


def direct_uptodate_deps_inverse_closure(job_id: CMJobID, db) -> Set[CMJobID]:
    """
        Closure of direct_uptodate_deps_inverse:
        all jobs that depend on this.
    """
    from compmake.jobs.queries import parents

    # all parents
    dep_inv = parents(job_id, db)
    # plus their definition closure
    from compmake.jobs.queries import definition_closure

    closure = definition_closure(dep_inv, db)
    # this is not true in general
    # assert not closure & dep_inv
    dep_inv.update(closure)
    # plus the ones that were defined by it
    from compmake.jobs.storage import get_job_cache

    if get_job_cache(job_id, db).state == Cache.DONE:
        dep_inv.update(jobs_defined(job_id, db))
    return dep_inv


class CacheQueryDB:
    """
        This works as a view on a DB which is assumed not to change
        between calls.
    """

    def __init__(self, db):
        self.db = db

    def invalidate(self):
        self.get_job_cache.reset()
        self.get_job.reset()
        self.all_jobs.reset()
        self.job_exists.reset()
        self.up_to_date.reset()
        self.direct_children.reset()
        self.direct_parents.reset()
        self.dependencies_up_to_date.reset()
        self.jobs_defined.reset()

    @memoized_reset
    def get_job_cache(self, job_id: CMJobID) -> Cache:
        from .storage import get_job_cache

        return get_job_cache(job_id, db=self.db)

    @memoized_reset
    def jobs_defined(self, job_id: CMJobID):
        return jobs_defined(job_id, db=self.db)

    @memoized_reset
    def get_job(self, job_id: CMJobID) -> Job:
        from .storage import get_job

        return get_job(job_id, db=self.db)

    @memoized_reset
    def all_jobs(self) -> List[CMJobID]:
        from .storage import all_jobs

        # NOTE: very important, do not memoize iterator
        res = list(all_jobs(db=self.db))
        return res

    @memoized_reset
    def job_exists(self, job_id: CMJobID) -> bool:
        from .storage import job_exists

        return job_exists(job_id=job_id, db=self.db)

    @memoized_reset
    def up_to_date(self, job_id: CMJobID) -> Tuple[bool, str, float]:
        with db_error_wrap("up_to_date()", job_id=job_id):
            return self._up_to_date_actual(job_id)

    def _up_to_date_actual(self, job_id: CMJobID) -> Tuple[bool, str, float]:
        with db_error_wrap("_up_to_date_actual()", job_id=job_id):
            cache = self.get_job_cache(job_id)  # OK

            if cache.state == Cache.NOT_STARTED:
                return False, "Not started", cache.timestamp

            if cache.timestamp == Cache.TIMESTAMP_TO_REMAKE:
                return False, "Marked invalid", cache.timestamp

            dependencies = self.direct_children(job_id)

            for child in dependencies:
                if not self.job_exists(child):
                    if CompmakeConstants.tolerate_db_inconsistencies:
                        logger.warn(f"Skipping not exiting child {child} of {job_id}")
                        # TODO: find out why
                        continue
                child_up, _, child_timestamp = self.up_to_date(child)
                if not child_up:
                    return False, f"At least: Dep {child!r} not up to date.", cache.timestamp
                else:
                    if child_timestamp > cache.timestamp:
                        return False, f"At least: Dep {child!r} have been updated.", cache.timestamp

            # plus jobs that defined it
            defined_by = list(self.get_job(job_id).defined_by)
            defined_by.remove(CMJobID("root"))
            dependencies.update(defined_by)

            for defby in defined_by:
                defby_up, _, _ = self.up_to_date(defby)
                if not defby_up:
                    return False, f"Definer {defby!r} not up to date.", cache.timestamp
                # don't check timestamp for definers

            # FIXME BUG if I start (in progress), children get updated,
            # I still finish the computation instead of starting again
            if cache.state == Cache.FAILED:
                return False, "Failed", cache.timestamp

            assert cache.state == Cache.DONE

            return True, "", cache.timestamp

    @memoized_reset
    def direct_children(self, job_id: CMJobID) -> Set[CMJobID]:
        from compmake.jobs.queries import direct_children

        return direct_children(job_id, db=self.db)

    @memoized_reset
    def direct_parents(self, job_id: CMJobID) -> Set[CMJobID]:
        from compmake.jobs.queries import direct_parents

        return direct_parents(job_id, db=self.db)

    @memoized_reset
    def parents(self, job_id: CMJobID) -> Set[CMJobID]:

        t = set()
        parents_jobs = self.direct_parents(job_id)
        for p in parents_jobs:
            t.add(p)
            t.update(self.parents(p))
        return t

    @memoized_reset
    def dependencies_up_to_date(self, job_id: CMJobID) -> bool:
        """ Returns true if all the dependencies are up to date """
        for child in self.direct_children(job_id):
            child_up, _, _ = self.up_to_date(child)
            if not child_up:
                return False
        return True

    def tree(self, jobs: Collection[CMJobID]) -> List[CMJobID]:
        """ More efficient version of tree()
            which is direct_children() recursively. """
        stack = []

        stack.extend(jobs)

        result = set()

        while stack:
            job_id = stack.pop()

            for c in self.direct_children(job_id):
                if not c in result:
                    result.add(c)
                    stack.append(c)

        return list(result)

    def list_todo_targets(self, jobs: Collection[CMJobID]) -> Tuple[Set[CMJobID], Set[CMJobID], Set[CMJobID]]:
        """
            Returns a tuple (todo, jobs_done, ready):
             todo:  set of job ids to do (children that are not up to date)
             done:  top level targets (in jobs) that are already done.
             ready: ready to do (dependencies_up_to_date)
        """
        with db_error_wrap("list_todo_targets()", jobs=jobs):
            for j in jobs:
                if not self.job_exists(j):
                    raise_desc(CompmakeBug, "Job does not exist", job_id=j)

            todo = set()
            done = set()
            seen = set()
            stack = list()
            stack.extend(jobs)

            class A:
                count = 0

            def summary():
                A.count += 1
                if A.count % 100 != 0:
                    return

            while stack:
                summary()

                job_id = stack.pop()
                seen.add(job_id)
                res = self.up_to_date(job_id)

                up, _, _ = res
                if up:
                    done.add(job_id)
                else:
                    todo.add(job_id)
                    for child in self.direct_children(job_id):
                        if not self.job_exists(child):
                            msg = f"Job {job_id!r} references a not existing job {child!r}. "
                            msg += (
                                "This might happen when you change a dynamic job "
                                "so that it changes the jobs it created. "
                                'Try "delete not root" to fix the DB.'
                            )
                            if CompmakeConstants.tolerate_db_inconsistencies:
                                logger.warn(msg)
                            else:
                                raise CompmakeBug(msg)
                        if not child in seen:
                            stack.append(child)

            todo_and_ready = set([job_id for job_id in todo if self.dependencies_up_to_date(job_id)])

            return todo, done, todo_and_ready

    # @contract(returns=set, jobs="unicode|set(unicode)")
    def tree_children_and_uodeps(self, jobs: Union[str, Set[str]]):
        """ Closure of the relation children and dependencies of userobject.
        """
        stack = []
        if isinstance(jobs, str):
            jobs = [jobs]

        stack.extend(jobs)

        result = set()

        def descendants(a_job_id):
            deps = collect_dependencies(get_job_userobject(a_job_id, self.db))
            children = self.direct_children(a_job_id)
            check_isinstance(children, set)
            r = children | deps
            check_isinstance(r, set)
            return r

        while stack:
            job_id = stack.pop()

            for c in descendants(job_id):
                if not self.job_exists(c):
                    raise ValueError(c)
                if not c in result:
                    result.add(c)
                    stack.append(c)

        return result


@contextmanager
def db_error_wrap(what, **args):
    try:
        yield
    except CompmakeDBError as e:
        raise_wrapped(CompmakeDBError, e, what, compact=True, **args)



import sys

from .helpers import COMMANDS_ADVANCED, GENERAL, ui_command
from .visualization import ui_info, ui_message
from .. import CompmakeGlobalState, get_compmake_config
from ..config import create_config_html, set_config_from_strings, show_config
from ..exceptions import UserError

__all__ = [
    "config",
    "config_html",
]


@ui_command(section=GENERAL)
def config(args, context):
    """ Get/set configuration parameters.

Usage:

    @: config  <switch>  <value>

Without arguments, shows all configuration switches.
 """
    if not args:
        # show
        b = show_config()
        ui_message(context, b)
        return

    name = args.pop(0)
    if not args:
        if not name in CompmakeGlobalState.config_switches:
            raise UserError(f"I don't know the switch '{name}'.")
        value = get_compmake_config(name)
        ui_info(context, f"config {name} {value}")
        return

    set_config_from_strings(name, args)


@ui_command(section=COMMANDS_ADVANCED)
def config_html(output_file=""):
    """ Dumps the config description in html on the specified file. """
    if output_file:
        f = open(output_file, "w")
    else:
        f = sys.stdout
    create_config_html(f)



from collections import OrderedDict, namedtuple

from compmake.jobs import CacheQueryDB, Cache
from compmake.jobs.storage import all_jobs
from compmake.ui import COMMANDS_ADVANCED, ui_command


@ui_command(section=COMMANDS_ADVANCED)
def gantt(job_list, context, filename="gantt.html"):
    """

    """
    from networkx.algorithms.dag import topological_sort

    db = context.get_compmake_db()
    if not job_list:
        #        job_list = list(top_targets(db))
        job_list = all_jobs(db)
    # plus all the jobs that were defined by them
    job_list = set(job_list)
    #    job_list.update(definition_closure(job_list, db))

    from networkx import DiGraph

    G = DiGraph()
    cq = CacheQueryDB(db)

    for job_id in job_list:
        cache = cq.get_job_cache(job_id)
        if cache.state != Cache.DONE:
            continue
        length = cache.int_make.get_cputime_used()
        attr_dict = dict(cache=cache, length=length)
        G.add_node(job_id, **attr_dict)

        dependencies = cq.direct_children(job_id)
        for c in dependencies:
            G.add_edge(c, job_id)

        defined = cq.jobs_defined(job_id)
        for c in defined:
            G.add_edge(job_id, c)

    order = list(topological_sort(G))
    for job_id in order:
        cache = cq.get_job_cache(job_id)
        if cache.state != Cache.DONE:
            continue
        length = G.nodes[job_id]["length"]
        pre = list(G.predecessors(job_id))
        #        print('%s pred %s' % (job_id, pre))
        if not pre:
            T0 = 0
            G.nodes[job_id]["CP"] = None
        else:
            # find predecessor with highest T1
            import numpy as np

            T1s = list(G.nodes[_]["T1"] for _ in pre)
            i = np.argmax(T1s)
            T0 = T1s[i]
            G.nodes[job_id]["CP"] = pre[i]
        T1 = T0 + length
        G.nodes[job_id]["T0"] = T0
        G.nodes[job_id]["T1"] = T1

        G.nodes[job_id]["critical"] = False

    sg_ideal = SimpleGantt()

    by_ideal_completion = sorted(order, key=lambda _: G.nodes[_]["T1"])
    last = by_ideal_completion[-1]
    path = []
    while last is not None:
        path.append(last)
        G.nodes[last]["critical"] = True
        last = G.nodes[last]["CP"]

    print("Critical path:")
    for job_id in reversed(path):
        length = G.nodes[job_id]["length"]
        print("-  %.1f s   %s" % (length, job_id))

    for job_id in by_ideal_completion:
        T0 = G.nodes[job_id]["T0"]
        T1 = G.nodes[job_id]["T1"]
        # length = G.nodes[job_id]['length']

        dependencies = list(G.predecessors(job_id))
        # cache = G.nodes[job_id]['cache']
        periods = OrderedDict()
        periods["ideal"] = (T0, T1)
        critical = G.nodes[job_id]["critical"]
        sg_ideal.add_job(job_id, dependencies, periods=periods, critical=critical)

    sg_actual = SimpleGantt()

    order_actual = sorted(order, key=lambda _: G.nodes[_]["cache"].int_make.t0)
    for job_id in order_actual:
        cache = G.nodes[job_id]["cache"]
        critical = G.nodes[job_id]["critical"]
        dependencies = list(G.predecessors(job_id))
        periods = OrderedDict()
        periods["make"] = cache.int_make.walltime_interval()
        sg_actual.add_job(job_id, dependencies, periods=periods, critical=critical)

    sg_actual_detailed = SimpleGantt()
    for job_id in order_actual:
        cache = G.nodes[job_id]["cache"]
        critical = G.nodes[job_id]["critical"]
        periods = OrderedDict()
        periods["load"] = cache.int_load_results.walltime_interval()
        periods["compute"] = cache.int_compute.walltime_interval()
        periods["gc"] = cache.int_gc.walltime_interval()
        periods["save"] = cache.int_save_results.walltime_interval()

        assert periods["load"][1] <= periods["compute"][0]
        assert periods["compute"][1] <= periods["save"][0]
        sg_actual_detailed.add_job(job_id, dependencies, periods=periods, critical=critical)

    html = ""
    width_pixels = 1000
    if True:
        html += "\n<h1>Actual</h1>"
        html += sg_actual.as_html(width_pixels)
    if True:
        html += "\n<h1>Actual (detailed)</h1>"
        html += sg_actual_detailed.as_html(width_pixels)
    if True:
        html += "\n<h1>Ideal</h1>"
        html += sg_ideal.as_html(width_pixels)

    html += (
        """
    <style>
        tr:hover td:first-child {

        }
        td:first-child {
        font-size: 10px;
        }
        td:nth-child(2) {
            background-color: grey;
            width: %spx;
        }
        .compute, .make, .save, .load, .ideal, .gc {
            outline: solid 1px black;
            float: left;
            clear: left;
        }
        .compute {
            background-color: red;
        }
        .make {
            background-color: blue;
        }
        .gc {
            background-color: brown;
        }
        .save {
            background-color: green;
        }
        .load {
            background-color: yellow;
        }
        .ideal {
            background-color: magenta;
        }
        .critical  {
            /* outline: solid 2px red !important; */
            background-color: pink;
        }
    </style>
        """
        % width_pixels
    )

    with open(filename, "w") as f:
        f.write(html)
    print("written to %s" % filename)


Entry = namedtuple("Entry", "dependencies periods critical")


class SimpleGantt(object):
    def __init__(self):
        self.entries = OrderedDict()

    def add_job(self, job_id, dependencies, periods, critical):
        E = Entry(dependencies, periods, critical)
        self.entries[job_id] = E

    def __str__(self):
        s = ""
        for job_id, e in self.entries.items():
            #            length = e.t1 - e.t0
            pass
        #            s += ('\n%40s  length %6.1f T0  %5.1f   T1  %5.1f' % (job_id[:40], length, e.t0, e.t1))
        return s

    def as_html(self, width_pixels):
        s = "<table>"
        abs_t0 = min(min(a[0] for a in _.periods.values()) for _ in self.entries.values())
        abs_t1 = max(max(a[1] for a in _.periods.values()) for _ in self.entries.values())

        def normalize_ts(x):
            return ((x - abs_t0) * 1.0 / (abs_t1 - abs_t0)) * width_pixels

        def normalize_length(L):
            return L / (abs_t1 - abs_t0) * width_pixels

        for job_id, e in self.entries.items():

            classes = ["critical"] if e.critical else []
            c = " ".join(classes)
            s += '\n<tr class="%s"><td>%s</td><td style="display: block;">' % (c, job_id[:20])
            for id_period, (t0, t1) in e.periods.items():
                r0 = normalize_ts(t0)
                w = normalize_length(t1 - t0)
                #                print('%s %10s %10s %10s w %s' % (job_id[:10], id_period, t0, t1, w))
                style = "display:block; margin-left: %spx; width: %spx; height: 10px" % (r0, w)
                s += "\n<span class='%s' style='%s'></span>" % (id_period, style)
            s += "\n</td></tr>"
        s += "\n</table>"

        return s



import os

from compmake.jobs.manager import Manager
from compmake.utils import isodate_with_secs
from .sge_misc import check_sge_environment
from .sge_sub import SGESub

__all__ = [
    "SGEManager",
]


class SGEManager(Manager):
    """ Runs compmake jobs using a SGE implementation """

    # @contract(num_processes=int, recurse="bool")
    def __init__(self, context, cq, recurse, num_processes):
        Manager.__init__(self, context=context, cq=cq, recurse=recurse)

        self.num_processes = num_processes

        check_sge_environment()

        storage = os.path.abspath(self.db.basepath)
        timestamp = isodate_with_secs().replace(":", "-")
        spool = os.path.join(storage, "sge", timestamp)
        if not os.path.exists(spool):
            os.makedirs(spool)

        self.sub_available = set()
        self.sub_processing = set()  # available + processing = subs.keys
        self.subs = {}  # name -> sub
        for i in range(self.num_processes):
            name = "w%02d" % i
            self.subs[name] = SGESub(name, db=self.db, spool=spool)
        self.job2subname = {}
        # all are available
        self.sub_available.update(self.subs)

    def get_resources_status(self):
        resource_available = {}

        assert len(self.sub_processing) == len(self.processing)
        # only one job at a time
        process_limit_ok = len(self.sub_processing) < self.num_processes
        if not process_limit_ok:
            resource_available["nproc"] = (False, "max %d nproc" % (self.num_processes))
            # this is enough to continue
            return resource_available
        else:
            resource_available["nproc"] = (True, "")

        return resource_available

    # @contract(reasons_why_not=dict)
    def can_accept_job(self, reasons_why_not):
        resources = self.get_resources_status()
        some_missing = False
        for k, v in resources.items():
            if not v[0]:
                some_missing = True
                reasons_why_not[k] = v[1]
        if some_missing:
            return False
        return True

    def instance_job(self, job_id):
        assert len(self.sub_available) > 0
        name = sorted(self.sub_available)[0]
        self.sub_available.remove(name)
        assert not name in self.sub_processing
        self.sub_processing.add(name)
        sub = self.subs[name]
        self.job2subname[job_id] = name
        ares = sub.instance_job(job_id)
        return ares

    def host_failed(self, job_id):
        Manager.host_failed(self, job_id)
        self._clear(job_id)

    def job_failed(self, job_id, deleted_jobs):
        Manager.job_failed(self, job_id, deleted_jobs)
        self._clear(job_id)

    def job_succeeded(self, job_id):
        Manager.job_succeeded(self, job_id)
        self._clear(job_id)

    def _clear(self, job_id):
        assert job_id in self.job2subname
        name = self.job2subname[job_id]
        del self.job2subname[job_id]
        assert name in self.sub_processing
        assert name not in self.sub_available
        self.sub_processing.remove(name)
        self.sub_available.add(name)

    def cleanup(self):
        Manager.cleanup(self)
        n = len(self.processing2result)
        if n > 100:
            print("Cleaning up %d SGE jobs. Please be patient." % n)
        for _, job in self.processing2result.items():
            job.delete_job()



# -*- coding: utf-8 -*-

import os
import signal
import warnings
from multiprocessing import Queue

from future.moves.queue import Empty

from compmake.events import broadcast_event, publish
from compmake.exceptions import MakeHostFailed
from compmake.jobs import Manager
from compmake.jobs.storage import get_job
from compmake.plugins.backend_multyvac.mvac_job_imp import mvac_job
from compmake.plugins.backend_multyvac.mvac_job_rdb_imp import mvac_job_rdb
from compmake.plugins.backend_pmake.parmake_job2_imp import parmake_job2
from compmake.plugins.backend_pmake.pmakesub import PmakeSub
from zuper_commons.fs import make_sure_dir_exists

__all__ = [
    "MVacManager",
]


class MVacManager(Manager):
    """
        Multyvac backend.
    """

    # @contract(num_processes="int")
    def __init__(
        self,
        context,
        cq,
        num_processes,
        recurse=False,
        show_output=False,
        new_process=False,
        volumes=[],
        rdb=False,
        rdb_vol=None,
        rdb_db=None,
    ):
        Manager.__init__(self, context=context, cq=cq, recurse=recurse)
        self.num_processes = num_processes
        self.last_accepted = 0
        self.cleaned = False
        self.show_output = show_output
        self.new_process = new_process
        self.volumes = volumes
        self.rdb = rdb
        self.rdb_db = rdb_db
        self.rdb_vol = rdb_vol

    def process_init(self):
        self.event_queue = Queue()
        self.event_queue_name = str(id(self))
        from compmake.plugins.backend_pmake.pmake_manager import PmakeManager

        PmakeManager.queues[self.event_queue_name] = self.event_queue

        # info('Starting %d processes' % self.num_processes)

        self.subs = {}  # name -> sub
        # available + processing + aborted = subs.keys
        self.sub_available = set()
        self.sub_processing = set()
        self.sub_aborted = set()

        self.signal_queue = Queue()

        db = self.context.get_compmake_db()
        storage = db.basepath  # XXX:
        logs = os.path.join(storage, "logs")
        for i in range(self.num_processes):
            name = "w%02d" % i
            write_log = os.path.join(logs, "%s.log" % name)
            make_sure_dir_exists(write_log)
            signal_token = name
            self.subs[name] = PmakeSub(
                name, signal_queue=self.signal_queue, signal_token=signal_token, write_log=write_log
            )
        self.job2subname = {}
        self.subname2job = {}
        # all are available at the beginning
        self.sub_available.update(self.subs)

        self.max_num_processing = self.num_processes

    def check_any_finished(self):
        # We make a copy because processing is updated during the loop
        try:
            token = self.signal_queue.get(block=False)
        except Empty:
            return False
        # print('received %r' % token)
        job_id = self.subname2job[token]
        self.subs[token].last
        self.check_job_finished(job_id, assume_ready=True)
        return True

    # XXX: boiler plate
    def get_resources_status(self):
        resource_available = {}

        assert len(self.sub_processing) == len(self.processing)

        if not self.sub_available:
            msg = "already %d nproc" % len(self.sub_processing)
            if self.sub_aborted:
                msg += " (%d workers aborted)" % len(self.sub_aborted)
            resource_available["nproc"] = (False, msg)
            # this is enough to continue
            return resource_available
        else:
            resource_available["nproc"] = (True, "")

        return resource_available

    # @contract(reasons_why_not=dict)
    def can_accept_job(self, reasons_why_not):
        if len(self.sub_available) == 0 and len(self.sub_processing) == 0:
            # all have failed
            msg = "All workers have aborted."
            raise MakeHostFailed(msg)

        resources = self.get_resources_status()
        some_missing = False
        for k, v in resources.items():
            if not v[0]:
                some_missing = True
                reasons_why_not[k] = v[1]
        if some_missing:
            return False
        return True

    def instance_job(self, job_id):
        publish(self.context, "worker-status", job_id=job_id, status="apply_async")
        assert len(self.sub_available) > 0
        name = sorted(self.sub_available)[0]
        self.sub_available.remove(name)
        assert not name in self.sub_processing
        self.sub_processing.add(name)
        sub = self.subs[name]

        self.job2subname[job_id] = name
        self.subname2job[name] = job_id

        job = get_job(job_id, self.db)

        if self.rdb:
            f = mvac_job_rdb
            args = (
                job_id,
                self.context,
                self.event_queue_name,
                self.show_output,
                self.volumes,
                self.rdb_vol.name,
                self.rdb_db,
                os.getcwd(),
            )
        else:
            if job.needs_context:
                # if self.new_process:
                #     f = parmake_job2_new_process
                #     args = (job_id, self.context)
                #
                # else:
                f = parmake_job2
                args = (job_id, self.context, self.event_queue_name, self.show_output)
            else:
                f = mvac_job
                args = (
                    job_id,
                    self.context,
                    self.event_queue_name,
                    self.show_output,
                    self.volumes,
                    os.getcwd(),
                )

        if True:
            async_result = sub.apply_async(f, args)
        else:
            warnings.warn("Debugging synchronously")
            async_result = f(args)

        return async_result

    def event_check(self):
        if not self.show_output:
            return
        while True:
            try:
                event = self.event_queue.get(block=False)  # @UndefinedVariable
                event.kwargs["remote"] = True
                broadcast_event(self.context, event)
            except Empty:
                break

    def process_finished(self):
        if self.cleaned:
            return
        self.cleaned = True

        # print('Clean up...')

        for name in self.sub_processing:
            self.subs[name].proc.terminate()

        for name in self.sub_available:
            self.subs[name].terminate()

        elegant = False
        # XXX: in practice this never works well
        if elegant:
            timeout = 1
            for name in self.sub_available:
                self.subs[name].proc.join(timeout)

        # XXX: ... so we just kill them mercilessly
        else:
            #  print('killing')
            for name in self.sub_processing:
                pid = self.subs[name].proc.pid
                os.kill(pid, signal.SIGKILL)

        self.event_queue.close()
        self.signal_queue.close()
        from compmake.plugins.backend_pmake.pmake_manager import PmakeManager

        del PmakeManager.queues[self.event_queue_name]

    # Normal outcomes
    def job_failed(self, job_id, deleted_jobs):
        Manager.job_failed(self, job_id, deleted_jobs)
        self._clear(job_id)

    def job_succeeded(self, job_id):
        Manager.job_succeeded(self, job_id)
        self._clear(job_id)

    def _clear(self, job_id):
        assert job_id in self.job2subname
        name = self.job2subname[job_id]
        del self.job2subname[job_id]
        del self.subname2job[name]
        assert name in self.sub_processing
        assert name not in self.sub_available
        self.sub_processing.remove(name)
        self.sub_available.add(name)

    def host_failed(self, job_id):
        Manager.host_failed(self, job_id)

        assert job_id in self.job2subname
        name = self.job2subname[job_id]
        del self.job2subname[job_id]
        del self.subname2job[name]
        assert name in self.sub_processing
        assert name not in self.sub_available
        self.sub_processing.remove(name)

        # put in sub_aborted
        self.sub_aborted.add(name)

    def cleanup(self):
        self.process_finished()



__version__ = "6.1.1"

from zuper_commons.logs import ZLogger

version = __version__

logger = ZLogger(__name__)

from .constants import *
from .state import *
from .storage import StorageFilesystem


from .structures import Promise
from .exceptions import *
from .context import Context
from .jobs import progress
from . import plugins
from .plugins.execution_stats import *
from .scripts.master import read_rc_files

# Default initialization
set_compmake_status(CompmakeConstants.compmake_status_embedded)

if CompmakeConstants.debug_check_invariants:
    logger.warn("debug_check_invariants = True: this might slow down quite a bit")



"""
    Main function:

        parse_job_list(tokens, context)

    Canonical forms:
        [A] except [B]     =>   A minus the elements in B
        [A] in [B]         =>   intersection of A and B

    Rewriting:
        not [job_list]     =>   $all except [job_list]
        except [job_list]  =>   $all except [job_list]

    Association:

        [A] except [B] except [C] == [A] except ([B] except [C])
        [A] in [B] in [C] == [A] in ([B] in [C])

    Priority:
        in > except > not



"""
import types
from collections import namedtuple

from compmake.constants import CompmakeConstants
from compmake.context import Context
from compmake.jobs.uptodate import CacheQueryDB
from zuper_commons.types import check_isinstance

from .. import get_job
from ...exceptions import CompmakeSyntaxError, UserError
from ...structures import Cache
from ...utils import expand_wildcard

__all__ = [
    "parse_job_list",
]

CompmakeConstants.aliases["last"] = "*"


def add_alias(alias, value):
    """ Sets the given alias to value. See eval_alias() for a discussion
    of the meaning of value. """
    CompmakeConstants.aliases[alias] = value


def assert_list_of_strings(l):
    assert all([isinstance(x, str) for x in l]), "Expected list of strings: %s." % str(l)


def is_alias(alias):
    return alias.lower() in CompmakeConstants.aliases


def eval_alias(alias, context: Context, cq: CacheQueryDB):
    """
    Evaluates the given alias.
    Returns a list of job_id strings.

    The value can have several types:
    - if it is a string, it is interpreted as a job id
    - if it is a list, it must be a list of string, interpreted as a job id
    - if it is callable (FunctionType),
      it is called, and it must return a list of strings.

    """

    alias = alias.lower()
    assert is_alias(alias)
    value = CompmakeConstants.aliases[alias]

    if isinstance(value, str):
        return list([value])
    elif isinstance(value, list):
        assert_list_of_strings(value)
        return value
    elif isinstance(value, types.FunctionType):
        result = value(context=context, cq=cq)
        # can be generator; no assert_list_of_strings(result)
        return result
    else:
        msg = 'I cannot interpret alias "%s" -> "%s".' % (alias, value)
        raise ValueError(msg)


def list_matching_functions(token: str, context: Context, cq: CacheQueryDB):
    db = context.get_compmake_db()
    assert token.endswith("()")
    if len(token) < 3:
        raise UserError('Malformed token "%s".' % token)

    function_id = token[:-2]

    num_matches = 0
    for job_id in cq.all_jobs():
        # command name (f.__name__)
        command_desc = get_job(job_id, db=db).command_desc
        if function_id.lower() == command_desc.lower():
            yield job_id
            num_matches += 1

    if num_matches == 0:
        raise UserError('Could not find matches for function "%s()".' % function_id)


def expand_job_list_token(token, context: Context, cq: CacheQueryDB):
    """ Parses a token (string). Returns a generator of jobs.
        Raises UserError, CompmakeSyntaxError """

    assert isinstance(token, str)

    if token.find("*") > -1:
        return expand_wildcard(token, cq.all_jobs())
    elif is_alias(token):
        return eval_alias(token, context, cq)
    elif token.endswith("()"):
        return list_matching_functions(token, context, cq)
    else:
        # interpret as a job id
        job_id = token
        if not cq.job_exists(job_id):
            msg = 'Job or expression "%s" not found.' % job_id
            raise UserError(msg)
        return [job_id]


def expand_job_list_tokens(tokens, context: Context, cq: CacheQueryDB):
    """ Expands a list of tokens using expand_job_list_token().
        yields job_id """
    for token in tokens:
        if not isinstance(token, str):
            # print tokens XXX
            pass
        for job in expand_job_list_token(token, context, cq):
            yield job


class Operators(object):
    Op = namedtuple("Op", "name")

    NOT = Op("not")
    DIFFERENCE = Op("difference")
    INTERSECTION = Op("intersection")

    translation = {
        "not": NOT,
        "except": DIFFERENCE,
        "but": DIFFERENCE,
        "in": INTERSECTION,
        "and": INTERSECTION,
        "intersect": INTERSECTION,
    }

    @staticmethod
    def parse(tokens):
        """ Parses a list of tokens for known operators.
        Returns a list where the operators are replaced by their codes. """

        def token2op(token):
            """ Translates one token, or returns the same """
            tokenl = token.lower()
            return Operators.translation.get(tokenl, token)

        return list(map(token2op, tokens))


def list_jobs_with_state(state, context: Context, cq: CacheQueryDB):
    """ Returns a list of jobs in the given state. """
    for job_id in cq.all_jobs():
        if cq.get_job_cache(job_id).state == state:  # TODO
            yield job_id


def list_ready_jobs(context: Context, cq: CacheQueryDB):
    """ Returns a list of jobs that can be done now,
        as their dependencies are up-to-date. """
    for job_id in cq.all_jobs():
        if cq.dependencies_up_to_date(job_id):
            yield job_id


def list_uptodate_jobs(context: Context, cq: CacheQueryDB):
    """ Returns a list of jobs that are uptodate
        (DONE, and all depednencies DONE)."""
    for job_id in cq.all_jobs():
        up, _, _ = cq.up_to_date(job_id)
        if up:
            yield job_id


def list_todo_jobs(context: Context, cq: CacheQueryDB):
    """
        Returns a list of jobs that haven't been DONE.
        Note that it could be DONE but not up-to-date.
    """
    for job_id in cq.all_jobs():
        if cq.get_job_cache(job_id).state != Cache.DONE:
            yield job_id


def list_root_jobs(context: Context, cq: CacheQueryDB):
    """ Returns a list of jobs that were defined by the original process.  """
    for job_id in cq.all_jobs():
        job = cq.get_job(job_id)
        if is_root_job(job):
            yield job_id


def list_generated_jobs(context: Context, cq: CacheQueryDB):
    """ Returns a list of jobs that were generated by other jobs.  """
    for job_id in cq.all_jobs():
        job = cq.get_job(job_id)
        if not is_root_job(job):
            yield job_id


def list_levelX_jobs(context, cq, X):
    """ Returns a list of jobs that are at level X """
    for job_id in cq.all_jobs():
        job = cq.get_job(job_id)
        level = len(job.defined_by) - 1
        if level == X:
            yield job_id


def list_level1_jobs(context: Context, cq: CacheQueryDB):
    for x in list_levelX_jobs(context, cq, 1):
        yield x


def list_level2_jobs(context: Context, cq: CacheQueryDB):
    for x in list_levelX_jobs(context, cq, 2):
        yield x


def list_level3_jobs(context: Context, cq: CacheQueryDB):
    for x in list_levelX_jobs(context, cq, 3):
        yield x


def list_level4_jobs(context: Context, cq: CacheQueryDB):
    for x in list_levelX_jobs(context, cq, 4):
        yield x


def is_root_job(job):
    return job.defined_by == ["root"]


def is_dynamic_job(job):
    return bool(job.needs_context)


def list_dynamic_jobs(context: Context, cq: CacheQueryDB):
    """ Returns a list of jobs that are uptodate
        (DONE, and all depednencies DONE)."""
    for job_id in cq.all_jobs():
        job = cq.get_job(job_id)
        if is_dynamic_job(job):
            yield job_id


def list_top_jobs(context: Context, cq: CacheQueryDB):
    """ Returns a list of jobs that are top-level targets.  """
    for job_id in cq.all_jobs():
        if not cq.direct_parents(job_id):
            yield job_id


def list_bottom_jobs(context: Context, cq: CacheQueryDB):
    """ Returns a list of jobs that do not depend on anything else. """
    for job_id in cq.all_jobs():
        if not cq.direct_children(job_id):  # TODO
            yield job_id


def obtain_all(context: Context, cq: CacheQueryDB):
    res = list(cq.all_jobs())
    print("obtain all: %s" % res)
    return res


add_alias("all", obtain_all)
add_alias("failed", lambda context, cq: list_jobs_with_state(Cache.FAILED, context=context, cq=cq))
add_alias("blocked", lambda context, cq: list_jobs_with_state(Cache.BLOCKED, context=context, cq=cq))
add_alias("ready", list_ready_jobs)
add_alias("todo", list_todo_jobs)
add_alias("top", list_top_jobs)
add_alias("uptodate", list_uptodate_jobs)
add_alias("root", list_root_jobs)
add_alias("generated", list_generated_jobs)
add_alias("level1", list_level1_jobs)
add_alias("level2", list_level2_jobs)
add_alias("level3", list_level3_jobs)
add_alias("level4", list_level4_jobs)
add_alias("dynamic", list_dynamic_jobs)
add_alias("bottom", list_bottom_jobs)
add_alias("done", lambda context, cq: list_jobs_with_state(Cache.DONE, context=context, cq=cq))
# add_alias('in_progress',
#           lambda context, cq:
#           list_jobs_with_state(Cache.IN_PROGRESS, context=context, cq=cq))
add_alias("not_started", lambda context, cq: list_jobs_with_state(Cache.NOT_STARTED, context=context, cq=cq))


def parse_job_list(tokens, context, cq=None):
    """
        Parses a job list. tokens can be:

        1. a string, in that case it is split()
        2. a list, in which case each element is treated as a token.

        NO(If tokens is not empty, then if it evaluates to empty,
        an error is raised (e.g. "make failed" and no failed jobs will
        throw an error).)

        Returns a list of strings.
    """
    if cq is None:
        cq = CacheQueryDB(context.get_compmake_db())

    if isinstance(tokens, str):
        tokens = tokens.strip().split()

    if not tokens:
        return []

    # First we look for operators
    ops = Operators.parse(tokens)

    # print(" %s => %s" % (tokens, ops))
    result = eval_ops(ops=ops, context=context, cq=cq)

    # FIXME, remove
    result = list(result)
    # print " %s => %s" % (tokens, result)

    return result


def eval_ops(ops, context: Context, cq: CacheQueryDB):
    """ Evaluates an expression.
      ops: list of strings and int representing operators """
    check_isinstance(ops, list)

    def list_split(l, index):
        """ Splits a list in two """
        return l[0:index], l[index + 1 :]

    # The sequence of the following operations
    # defines the associativity rules

    # in > except > not

    if Operators.INTERSECTION in ops:
        left, right = list_split(ops, ops.index(Operators.INTERSECTION))
        if not left or not right:
            msg = """ INTERSECTION requires only a right argument.
            Interpreting "%s" INTERSECTION "%s". """ % (
                " ".join(left),
                " ".join(right),
            )
            raise CompmakeSyntaxError(msg)
        left = eval_ops(ops=left, context=context, cq=cq)
        right = set(eval_ops(ops=right, context=context, cq=cq))
        for x in left:
            if x in right:
                yield x

    elif Operators.DIFFERENCE in ops:
        left, right = list_split(ops, ops.index(Operators.DIFFERENCE))
        if not left or not right:
            msg = """ EXCEPT requires a left and right argument.
            Interpreting "%s" EXCEPT "%s". """ % (
                " ".join(left),
                " ".join(right),
            )
            raise CompmakeSyntaxError(msg)

        left = eval_ops(ops=left, context=context, cq=cq)
        right = set(eval_ops(ops=right, context=context, cq=cq))
        for x in left:
            if x not in right:
                yield x

    elif Operators.NOT in ops:
        left, right = list_split(ops, ops.index(Operators.NOT))
        if left or not right:  # forbid left, require right
            msg = """ NOT requires only a right argument. Interpreting "%s" NOT
                    "%s". """ % (
                " ".join(left),
                " ".join(right),
            )
            raise CompmakeSyntaxError(msg)

        right_res = set(eval_ops(ops=right, context=context, cq=cq))
        # if not all_jobs:
        # assert False
        # print("NOT")
        #         print(' all_jobs evalatued to %r' % (all_jobs))
        #         print(' right ops %r evalatued to %r' % (right, right_res))
        #         result = []
        for x in cq.all_jobs():
            if not x in right_res:
                yield x
                #
                #             in_right = x in right_res
                #             print('   is %r in not set -> %s' % (x,
                # in_right))
                #             if not in_right:
                #                 result.append(x)
                #         print(' result -> %s' % result)
                #         for x in result:
                #             yield x

    else:
        # no operators: simple list
        # cannot do this anymore, now it's a generator.
        # assert_list_of_strings(ops)
        for x in expand_job_list_tokens(ops, context=context, cq=cq):
            yield x



# -*- coding: utf-8 -*-

import os

from compmake.context import Context
from compmake.exceptions import CompmakeBug, HostFailed, JobFailed
from compmake.jobs import (
    get_job_args,
    job2cachekey,
    job2jobargskey,
    job2key,
    job2userobjectkey,
    result_dict_check,
)
from compmake.state import get_compmake_config
from compmake.storage.filesystem import StorageFilesystem
from zuper_commons.types import check_isinstance
from .logging_imp import disable_logging_if_config

__all__ = [
    "mvac_job_rdb",
]


def mvac_job_rdb_instance(context, job_id, volumes, rdb_vol_name, rdb_db, cwd):
    import multyvac

    layer = get_compmake_config("multyvac_layer")
    if not layer:
        layer = None
    all_volumes = volumes + [rdb_vol_name]

    command, _, _ = get_job_args(job_id, db=context.get_compmake_db())
    misc = dict(deps=[command])

    # print('Instancing (volumes: %r, layer=%r)' % (all_volumes, layer))
    core = get_compmake_config("multyvac_core")
    multyvac_job_id = multyvac.submit(
        mvac_job_rdb_worker,
        job_id=job_id,
        rdb_basepath=rdb_db.basepath,
        misc=misc,
        cwd=cwd,
        _core=core,
        _name=job_id,
        _layer=layer,
        _vol=all_volumes,
    )
    # print('Getting job %r' % multyvac_job_id)
    multyvac_job = multyvac.get(multyvac_job_id)
    # print('Got job')
    return multyvac_job


def mvac_job_rdb_worker(job_id, rdb_basepath, cwd, misc):
    from compmake.jobs.actions import make

    rdb = StorageFilesystem(rdb_basepath)
    context = Context(rdb)

    if not os.path.exists(cwd):
        print("cwd %r not existing", cwd)
        os.makedirs(cwd)
    os.chdir(cwd)

    try:
        res = make(job_id, context=context)
    except JobFailed as e:
        res = e.get_result_dict()
    except HostFailed as e:
        res = e.get_result_dict()
    except CompmakeBug as e:
        res = e.get_result_dict()
    except Exception as e:
        res = CompmakeBug(str(e)).get_result_dict()

    result_dict_check(res)
    print("res: %r" % res)
    return res


# @contract(args="tuple(str, *,  str, bool, list, *, *, str)")
def mvac_job_rdb(args):
    import multyvac

    job_id, context, event_queue_name, show_output, volumes, rdb_vol_name, rdb_db, cwd = args

    check_isinstance(job_id, str)

    # Disable multyvac logging
    disable_logging_if_config(context)

    multyvac_job = mvac_job_rdb_instance(context, job_id, volumes, rdb_vol_name, rdb_db, cwd)
    multyvac_job.wait()

    db = context.get_compmake_db()
    vol = multyvac.volume.get(rdb_vol_name)  # @UndefinedVariable

    res = multyvac_job.get_result()
    result_dict_check(res)

    # is there something to download?
    stuff = ("fail" in res) or ("new_jobs" in res)
    # alternatives = bug, abort
    if stuff:
        new_jobs = res.get("new_jobs", [])
        transfer_down(db, vol, rdb_db, job_id, new_jobs)
    return res


def transfer_down(db, vol, rdb_db, job_id, new_jobs):
    keys = get_keys_to_download(job_id, new_jobs, results=False)
    for key in keys:
        fr = rdb_db.filename_for_key(key)
        local_path = db.filename_for_key(key)
        remote_path = os.path.relpath(fr, rdb_db.basepath)
        # print('down %r->%r' % (remote_path, local_path))
        vol.get_file(remote_path, local_path)


def get_keys_to_download(job_id, new_jobs, results=False):
    keys = []

    # result of the job
    keys.append(job2cachekey(job_id))

    if results:
        keys.append(job2userobjectkey(job_id))

    for job_id in new_jobs:
        for r in [job2jobargskey, job2key]:
            key = r(job_id)
            keys.append(key)

    return keys



import os
import sys
from typing import List, Optional, Set, TYPE_CHECKING, Union

__all__ = [
    "Context",
]

if TYPE_CHECKING:
    from .storage import StorageFilesystem


class Context:
    # @contract(db='None|unicode|isinstance(StorageFilesystem)',
    #           currently_executing='None|list(unicode)')
    def __init__(
        self,
        db: "Optional[Union[str, StorageFilesystem]]" = None,
        currently_executing: Optional[List[str]] = None,
    ):
        """
            db: if a string, it is used as path for the DB

            currently_executing: str, job currently executing
                defaults to ['root']
        """
        if currently_executing is None:
            currently_executing = ["root"]
        from compmake import StorageFilesystem

        if db is None:
            prog, _ = os.path.splitext(os.path.basename(sys.argv[0]))

            # logger.info('Context(): Using default storage dir %r.' % prog)
            dirname = "out-%s" % prog
            db = StorageFilesystem(dirname, compress=True)

        if isinstance(db, str):
            db = StorageFilesystem(db, compress=True)

        assert db is not None
        self.compmake_db = db
        self._jobs_defined_in_this_session = set()
        self.currently_executing = currently_executing
        self._job_prefix = None

        # RC files read
        self.rc_files_read = []

        # counters for prefixes (generate_job_id)
        self.generate_job_id_counters = {}

    # This is used to make sure that the user doesn't define the same job
    # twice.
    def was_job_defined_in_this_session(self, job_id: str) -> bool:
        return job_id in self._jobs_defined_in_this_session

    def add_job_defined_in_this_session(self, job_id: str) -> None:
        self._jobs_defined_in_this_session.add(job_id)

    def get_jobs_defined_in_this_session(self) -> Set[str]:
        return set(self._jobs_defined_in_this_session)

    def reset_jobs_defined_in_this_session(self, jobs):
        """ Called only when initializing the context. """
        self._jobs_defined_in_this_session = set(jobs)

    def get_compmake_db(self):
        return self.compmake_db

    def get_comp_prefix(self):
        return self._job_prefix

    def comp_prefix(self, prefix):
        if prefix is not None:
            if " " in prefix:
                msg = "Invalid job prefix %r." % prefix
                from .exceptions import UserError

                raise UserError(msg)

        self._job_prefix = prefix

    # setting up jobs
    def comp_dynamic(self, command_, *args, **kwargs):
        from compmake.ui.ui import comp_

        return comp_(self, command_, *args, needs_context=True, **kwargs)

    def comp(self, command_, *args, **kwargs):
        from compmake.ui.ui import comp_

        return comp_(self, command_, *args, **kwargs)

    def comp_store(self, x, job_id=None):
        return comp_store_(x=x, context=self, job_id=job_id)

    def interpret_commands_wrap(self, commands):
        """
            Returns:

            0            everything ok
            int not 0    error
            string       an error, explained

            False?       we want to exit (not found in source though)
        """
        from .ui import interpret_commands_wrap
        from .jobs import CacheQueryDB

        cq = CacheQueryDB(self.get_compmake_db())
        return interpret_commands_wrap(commands, context=self, cq=cq)

    def batch_command(self, s) -> None:
        from .ui import batch_command
        from .jobs import CacheQueryDB

        cq = CacheQueryDB(self.get_compmake_db())
        return batch_command(s, context=self, cq=cq)

    def compmake_console(self):

        from .ui import compmake_console_text

        compmake_console_text(self)


def comp_store_(x, context, job_id=None):
    """

    Stores the object as a job, keeping track of whether
        we have it.
    """

    id_object = id(x)

    book = context.comp_store.objectid2job
    if id_object not in book:
        job_params = {}
        if job_id is not None:
            job_params["job_id"] = job_id

        job = context.comp(load_static_storage, x, **job_params)
        book[id_object] = job
    return book[id_object]


def load_static_storage(x):  # XXX: this uses double the memory though
    return x



import os
import unittest
from abc import ABCMeta
from shutil import rmtree
from tempfile import mkdtemp

from compmake import logger, set_compmake_config
from compmake.context import Context
from compmake.exceptions import CommandFailed, MakeFailed
from compmake.jobs import get_job, parse_job_list
from compmake.scripts.master import compmake_main
from compmake.storage import StorageFilesystem
from compmake.structures import CMJobID, Job


class CompmakeTest(unittest.TestCase):
    __metaclass__ = ABCMeta

    def setUp(self):
        self.root0 = mkdtemp()
        self.root = os.path.join(self.root0, "compmake")
        self.db = StorageFilesystem(self.root, compress=True)
        self.cc = Context(db=self.db)
        # don't use '\r'
        set_compmake_config("interactive", False)
        set_compmake_config("console_status", False)

        from compmake.constants import CompmakeConstants

        CompmakeConstants.debug_check_invariants = True
        self.mySetUp()

    def tearDown(self):
        if True:
            print("not deleting %s" % self.root0)
        else:
            rmtree(self.root0)
        from multiprocessing import active_children

        c = active_children()
        print("active children: %s" % c)
        if c:
            if True:
                msg = "Still active children: %s" % c
                logger.warning(msg)
            else:
                raise Exception(msg)

    # optional init
    # noinspection PyPep8Naming
    def mySetUp(self):
        pass

    # useful
    def comp(self, *args, **kwargs):
        return self.cc.comp(*args, **kwargs)

    # @contract(job_id="unicode", returns=Job)
    def get_job(self, job_id: CMJobID) -> Job:
        db = self.cc.get_compmake_db()
        return get_job(job_id=job_id, db=db)

    def get_jobs(self, expression):
        """ Returns the list of jobs corresponding to the given expression. """
        return list(parse_job_list(expression, context=self.cc))

    def assert_cmd_success(self, cmds):
        """ Executes the (list of) commands and checks it was succesful. """
        print("@ %s" % cmds)
        try:
            self.cc.batch_command(cmds)
        except MakeFailed as e:
            print("Detected MakeFailed")
            print("Failed jobs: %s" % e.failed)
            for job_id in e.failed:
                self.cc.interpret_commands_wrap("details %s" % job_id)

        except CommandFailed:
            # msg = 'Command %r failed. (res=%s)' % (cmds, res)
            raise

        self.cc.interpret_commands_wrap("check_consistency raise_if_error=1")

    def assert_cmd_fail(self, cmds):
        """ Executes the (list of) commands and checks it was succesful. """
        print("@ %s     [supposed to fail]" % cmds)
        try:
            self.cc.batch_command(cmds)
        except CommandFailed:
            pass
        else:
            msg = "Command %r did not fail." % cmds
            raise Exception(msg)

    def assert_cmd_success_script(self, cmd_string: str):
        """ This runs the "compmake_main" script which recreates the DB and
        context from disk. """
        ret = compmake_main([self.root, "--nosysexit", "-c", cmd_string])
        self.assertEqual(ret, 0)

    # useful mcdp_lang_tests
    def assert_defined_by(self, job_id, expected):
        self.assertEqual(self.get_job(job_id).defined_by, expected)

    def assertEqualSet(self, a, b):
        self.assertEqual(set(a), set(b))

    # @contract(expr="unicode")
    def assertJobsEqual(self, expr: str, jobs, ignore_dyn_reports=True):

        # js = 'not-valid-yet'
        js = self.get_jobs(expr)
        if ignore_dyn_reports:
            js = [x for x in js if not "dynreports" in x]
        try:
            self.assertEqualSet(js, jobs)
        except:
            print("expr %r -> %s" % (expr, js))
            print("differs from %s" % jobs)
            raise

    def assertMakeFailed(self, func, nfailed, nblocked):
        try:
            func()
        except MakeFailed as e:
            if len(e.failed) != nfailed:
                msg = "Expected %d failed, got %d: %s" % (nfailed, len(e.failed), e.failed)
                raise Exception(msg)
            if len(e.blocked) != nblocked:
                msg = "Expected %d blocked, got %d: %s" % (nblocked, len(e.blocked), e.blocked)
                raise Exception(msg)
        except Exception as e:
            raise Exception("unexpected: %s" % e)

    def assert_job_uptodate(self, job_id, status):
        res = self.up_to_date(job_id)
        self.assertEqual(res, status, "Want %r uptodate? %s" % (job_id, status))

    # @contract(returns=bool)
    def up_to_date(self, job_id) -> bool:
        from compmake.jobs.uptodate import CacheQueryDB

        cq = CacheQueryDB(db=self.db)
        up, reason, timestamp = cq.up_to_date(job_id)
        print("up_to_date(%r): %s, %r, %s" % (job_id, up, reason, timestamp))
        return up



from compmake.jobs.uptodate import direct_uptodate_deps_inverse_closure
from compmake.jobs.queries import definition_closure, jobs_defined
from compmake.jobs.uptodate import direct_uptodate_deps_inverse

from compmake.unittests.compmake_test import CompmakeTest
from nose.tools import istest


def always():
    pass


def other():
    pass


def fd(context):
    context.comp(always)
    print("fd sees %s" % TestDynamic8.define_other)
    if TestDynamic8.define_other:
        context.comp(other)


def mockup8(context):
    context.comp_dynamic(fd)


@istest
class TestDynamic8(CompmakeTest):

    define_other = True

    def test_dynamic8_remake(self):
        #         """ Re-execution creates more jobs.  """
        mockup8(self.cc)
        # run it
        TestDynamic8.define_other = True
        self.assert_cmd_success("make recurse=1")
        # we have three jobs defined
        self.assertJobsEqual("all", ["fd", "fd-always", "fd-other"])
        # clean and remake fd
        TestDynamic8.define_other = False
        self.assert_cmd_success("remake fd")
        # now the "other" job should disappear
        self.assertJobsEqual("all", ["fd", "fd-always"])

    def test_dynamic8_clean(self):
        #         """ Re-execution creates more jobs.  """
        mockup8(self.cc)
        # run it
        TestDynamic8.define_other = True
        self.assert_cmd_success("make recurse=1")
        # we have three jobs defined
        self.assertJobsEqual("all", ["fd", "fd-always", "fd-other"])
        # clean and remake fd
        TestDynamic8.define_other = False

        self.assertJobsEqual("done", ["fd", "fd-always", "fd-other"])
        self.assertEqualSet(jobs_defined("fd", self.db), ["fd-always", "fd-other"])

        self.assertEqualSet(definition_closure(["fd"], self.db), ["fd-always", "fd-other"])
        direct = direct_uptodate_deps_inverse("fd", self.db)
        self.assertEqualSet(direct, ["fd-always", "fd-other"])
        direct_closure = direct_uptodate_deps_inverse_closure("fd", self.db)
        self.assertEqualSet(direct_closure, ["fd-always", "fd-other"])

        self.assert_cmd_success("clean fd")
        # clean should get rid of the jobs
        self.assertJobsEqual("all", ["fd"])
        self.assert_cmd_success("make fd")
        # now the "other" job should disappear
        self.assertJobsEqual("all", ["fd", "fd-always"])

    def test_dynamic8_inverse(self):
        """ Re-execution creates fewer jobs. """
        mockup8(self.cc)
        # run it
        TestDynamic8.define_other = False
        self.assert_cmd_success("make recurse=1")
        # we have two jobs defined
        self.assertJobsEqual("all", ["fd", "fd-always"])
        # clean and remake fd
        TestDynamic8.define_other = True
        self.assert_cmd_success("remake fd")
        # now the "other" job should disappear
        self.assertJobsEqual("all", ["fd", "fd-always", "fd-other"])
        self.assertJobsEqual("done", ["fd", "fd-always"])



""" This plugin dumps all events received. """
import sys

from ..events import register_fallback_handler
from ..ui import compmake_colored
from ..utils import pad_to_screen

# We save it, because it will be redirected during job execution
stream = sys.stderr
other_stream = sys.stdout


def print_event(context, event):
    other_stream.flush()

    # age = time.time() - event.timestamp
    #    if age > 0.5:
    #        ages = '%.3fs ago' % age
    #    else:
    #        ages = ""

    s = str(event.kwargs)
    #    print ('%r has len %d' % (s, len(s)))
    MAX_LEN = 1000  # TODO:
    # TODO: clip_to_length(s, ' [...]')
    if len(s) > MAX_LEN:
        suff = " [...]"
        s = s[: MAX_LEN - len(suff)] + suff
    #        s = s[:MAX_LEN]

    msg = f"{event.name}: {s}"
    msg = compmake_colored(pad_to_screen(msg), "yellow")
    stream.write(msg)
    stream.write("\n")
    stream.flush()


if False:
    register_handler("*", print_event)
else:
    register_fallback_handler(print_event)



import logging
import traceback
from logging import Formatter
from time import time

from compmake import get_compmake_config, List, logger
from compmake.events import publish
from compmake.exceptions import JobFailed, JobInterrupted
from compmake.structures import Cache, CMJobID, IntervalTimer
from compmake.utils import OutputCapture, setproctitle
from zuper_commons.types import check_isinstance

from .dependencies import collect_dependencies
from .job_execution import job_compute
from .progress_imp2 import init_progress_tracking
from .queries import definition_closure, direct_parents
from .storage import (
    delete_all_job_data,
    delete_job_cache,
    get_job,
    get_job_cache,
    job_cache_exists,
    job_exists,
    set_job,
    set_job_cache,
    set_job_userobject,
)
from .uptodate import CacheQueryDB


def clean_targets(job_list: List[CMJobID], db, cq: CacheQueryDB):
    #     print('clean_targets (%r)' % job_list)
    job_list = set(job_list)

    # now we need to delete the definition closure
    # logger.info('getting closure')
    closure = definition_closure(job_list, db)

    basic = job_list - closure

    # logger.info(job_list=job_list, closure=closure, basic=basic)
    other_clean = set()
    for job_id in job_list:
        other_clean.update(cq.parents(job_id))

    other_clean -= closure
    #
    #     print('deleting: %r' % closure)
    #     print('only cleaning: %r' % basic)
    #     print('other cleaning: %r' % other_clean)
    #

    ccr = closure | basic | other_clean

    # logger.info(job_list=job_list, closure=closure, ccr=ccr)
    for job_id in ccr:
        # logger.info('clean_cache_relations', job_id=job_id)
        clean_cache_relations(job_id, db)

    # delete all in closure
    for job_id in closure:
        # logger.info('delete_all_job_data', job_id=job_id)
        delete_all_job_data(job_id, db)

    # just remove cache in basic
    for job_id in basic:
        # Cleans associated objects
        if job_cache_exists(job_id, db):
            # logger.info('delete_job_cache', job_id=job_id)
            delete_job_cache(job_id, db)
    # logger.info('done')
    # now we have to undo this one:
    # jobs_depending_on_this = direct_parents(job_id, self.db)
    # deps = result['user_object_deps']
    # for parent in jobs_depending_on_this:
    #     db_job_add_dynamic_children(job_id=parent, children=deps,
    #                                 returned_by=job_id, db=self.db)
    #     for d in deps:
    #         db_job_add_parent(job_id=d, parent=parent, db=self.db)


def clean_cache_relations(job_id: CMJobID, db):
    # print('cleaning cache relations for %r ' % job_id)
    if not job_exists(job_id, db):
        logger.warning("Cleaning cache for job %r which does not exist anymore; ignoring" % job_id)
        return

    # for all jobs that were done
    cache = get_job_cache(job_id, db)
    if cache.state == Cache.DONE:
        for parent in direct_parents(job_id, db):
            if not job_exists(parent, db):
                msg = (
                    "Could not find job %r (parent of %s) - ok if the job was deleted"
                    " otherwise it is a bug" % (parent, job_id)
                )
                logger.warning(msg)
                continue
            parent_job = get_job(parent, db)
            # print('  parent %r has dynamic %s' % (parent, parent_job.dynamic_children))
            if not job_id in parent_job.dynamic_children:
                # print('    skipping parent %r ' % parent)
                continue
            else:
                dynamic_children = parent_job.dynamic_children[job_id]
                # print('    dynamic_children %s' % parent_job.dynamic_children)
                # print('    children %s' % parent_job.children)
                del parent_job.dynamic_children[job_id]
                parent_job.children = parent_job.children - dynamic_children
                set_job(parent, parent_job, db)
                # print('     changed in %s' % parent_job.children)


def mark_to_remake(job_id: CMJobID, db):
    """ Delets and invalidates the cache for this object """
    # TODO: think of the difference between this and clean_target
    cache = get_job_cache(job_id, db)
    if cache.state == Cache.DONE:
        cache.timestamp = Cache.TIMESTAMP_TO_REMAKE
    set_job_cache(job_id, cache, db=db)


def mark_as_blocked(job_id: CMJobID, dependency=None, db=None) -> None:  # XXX
    cache = Cache(Cache.BLOCKED)
    cache.exception = f"Failure of dependency {dependency!r}"
    cache.backtrace = ""
    set_job_cache(job_id, cache, db=db)


def mark_as_failed(job_id: CMJobID, exception=None, backtrace=None, db=None) -> None:
    """ Marks job_id  as failed """
    cache = Cache(Cache.FAILED)
    if isinstance(exception, str):
        pass
    else:
        exception = exception.__str__()

    check_isinstance(backtrace, (type(None), str))
    cache.exception = exception
    cache.backtrace = backtrace
    cache.timestamp = time()
    set_job_cache(job_id, cache, db=db)


def make(job_id: CMJobID, context, echo=False):
    """
        Makes a single job.

        Returns a dictionary with fields:
             "user_object"
             "user_object_deps" = set of Promises
             "new_jobs" -> new jobs defined
             "deleted_jobs" -> jobs that were defined but not anymore

        Raises JobFailed
        or JobInterrupted. Also SystemExit, KeyboardInterrupt, MemoryError are
        captured.
    """
    db = context.get_compmake_db()

    int_make = IntervalTimer()

    host = "hostname"  # XXX

    if get_compmake_config("set_proc_title"):
        setproctitle(f"cm-{job_id}")

    # TODO: should we make sure we are up to date???
    #     up, reason = up_to_date(job_id, db=db)
    #     if up:
    #         msg = 'Job %r appears already done.' % job_id
    #         msg += 'This can only happen if another compmake process uses the ' \
    #                'same DB.'
    # logger.error(msg)
    #         user_object = get_job_userobject(job_id, db=db)
    #         # XXX: this is not right anyway
    #         return dict(user_object=user_object,
    #                     user_object_deps=collect_dependencies(user_object),
    #                     deleted_jobs=[],
    #                     new_jobs=[])

    job = get_job(job_id, db=db)
    cache = get_job_cache(job_id, db=db)

    if cache.state == Cache.DONE:
        prev_defined_jobs = set(cache.jobs_defined)
        # print('%s had previously defined %s' % (job_id, prev_defined_jobs))
    else:
        # print('%s was not DONE' % job_id)
        prev_defined_jobs = None

    # Note that at this point we save important information in the Cache
    # so if we set this then it's going to destroy it
    # cache.state = Cache.IN _ PROGRESS
    # set_job_cache(job_id, cache, db=db)

    # TODO: delete previous user object

    def progress_callback(stack):
        publish(context, "job-progress-plus", job_id=job_id, host=host, stack=stack)

    init_progress_tracking(progress_callback)

    disable_capture = False
    if disable_capture:
        logger.warning("Capture is disabled")
    if disable_capture:
        capture = None
    else:
        echo = False
        capture = OutputCapture(
            context=context,
            prefix=job_id,
            # This is instantaneous echo and should be False
            # They will generate events anyway.
            echo_stdout=echo,
            echo_stderr=echo,
        )

    # TODO: add whether we should just capture and not echo
    old_emit = logging.StreamHandler.emit

    from compmake.ui.coloredlog import colorize_loglevel

    FORMAT = "%(name)10s|%(filename)15s:%(lineno)-4s - %(funcName)-15s| %(message)s"

    formatter = Formatter(FORMAT)

    class Store(object):
        nhidden = 0

    def my_emit(_, log_record):
        # note that log_record.msg might be an exception
        try:
            try:
                s = str(log_record.msg)

            except:
                s = f"Could not print log_record {id(log_record)}"
            log_record.msg = colorize_loglevel(log_record.levelno, s)
            res = formatter.format(log_record)
            print(res)
            # this will be captured by OutputCapture anyway
        except:
            Store.nhidden += 1

    logging.StreamHandler.emit = my_emit

    already = set(context.get_jobs_defined_in_this_session())

    def get_deleted_jobs():
        generated = set(context.get_jobs_defined_in_this_session()) - already
        # print('failure: rolling back %s' % generated)

        from compmake.ui.ui import delete_jobs_recurse_definition

        todelete_ = set()
        # delete the jobs that were previously defined
        if prev_defined_jobs:
            todelete_.update(prev_defined_jobs)
        # and also the ones that were generated
        todelete_.update(generated)

        deleted_jobs_ = delete_jobs_recurse_definition(jobs=todelete_, db=db)
        # now we failed, so we need to roll back other changes
        # to the db
        return deleted_jobs_

    try:
        result = job_compute(job=job, context=context)

        assert isinstance(result, dict) and len(result) == 5
        user_object = result["user_object"]
        new_jobs = result["new_jobs"]
        int_load_results = result["int_load_results"]
        int_compute = result["int_compute"]
        int_gc = result["int_gc"]
        int_gc.stop()

    except KeyboardInterrupt as e:
        bt = traceback.format_exc()
        deleted_jobs = get_deleted_jobs()
        mark_as_failed(job_id, "KeyboardInterrupt: " + str(e), backtrace=bt, db=db)

        cache = get_job_cache(job_id, db=db)
        if capture is not None:
            cache.captured_stderr = capture.get_logged_stderr()
            cache.captured_stdout = capture.get_logged_stdout()
        else:
            msg = "(Capture turned off.)"
            cache.captured_stderr = msg
            cache.captured_stdout = msg

        set_job_cache(job_id, cache, db=db)

        raise JobInterrupted(job_id=job_id, deleted_jobs=list(deleted_jobs))

    except (
        BaseException,
        ArithmeticError,
        BufferError,
        LookupError,
        Exception,
        SystemExit,
        MemoryError,
    ) as e:
        bt = traceback.format_exc()
        s = "%s: %s" % (type(e).__name__, e)
        mark_as_failed(job_id, s, backtrace=bt, db=db)
        deleted_jobs = get_deleted_jobs()

        cache = get_job_cache(job_id, db=db)
        if capture is not None:
            cache.captured_stderr = capture.get_logged_stderr()
            cache.captured_stdout = capture.get_logged_stdout()
        else:
            msg = "(Capture turned off.)"
            cache.captured_stderr = msg
            cache.captured_stdout = msg

        set_job_cache(job_id, cache, db=db)

        raise JobFailed(job_id=job_id, reason=s, bt=bt, deleted_jobs=list(deleted_jobs))
    finally:
        int_finally = IntervalTimer()
        if capture is not None:
            capture.deactivate()
        # even if we send an error, let's save the output of the process
        logging.StreamHandler.emit = old_emit
        if Store.nhidden > 0:
            msg = "compmake: There were %d messages hidden due to bugs in logging." % Store.nhidden
            print(msg)
        int_finally.stop()
    #        print('finally: %s' % int_finally)

    int_save_results = IntervalTimer()

    # print('Now %s has defined %s' % (job_id, new_jobs))
    if prev_defined_jobs is not None:
        # did we defined fewer jobs this time around?
        # then we need to delete them
        todelete = set()
        for x in prev_defined_jobs:
            if x not in new_jobs:
                todelete.add(x)
        from compmake.ui.ui import delete_jobs_recurse_definition

        deleted_jobs = delete_jobs_recurse_definition(jobs=todelete, db=db)
    else:
        deleted_jobs = set()

    # print('Now %s has deleted %s' % (job_id, deleted_jobs))

    set_job_userobject(job_id, user_object, db=db)
    int_save_results.stop()

    #    logger.debug('Save time for %s: %s s' % (job_id, walltime_save_result))

    int_make.stop()
    end_time = time()

    cache = Cache(Cache.DONE)

    #    print('int_make: %s' % int_make)
    #    print('int_load_results: %s' % int_load_results)
    #    print('int_compute: %s' % int_compute)
    if int_gc.get_walltime_used() > 1.0:
        logger.warning("Expensive garbage collection detected at the end of %s: %s" % (job_id, int_gc))
    #    print('int_save_results: %s' % int_save_results)

    cache.int_make = int_make
    cache.int_load_results = int_load_results
    cache.int_compute = int_compute
    cache.int_gc = int_gc
    cache.int_save_results = int_save_results

    cache.timestamp = end_time

    cache.walltime_used = int_make.get_walltime_used()
    cache.cputime_used = int_make.get_cputime_used()
    cache.host = host
    cache.jobs_defined = new_jobs
    set_job_cache(job_id, cache, db=db)

    return dict(
        user_object=user_object,
        user_object_deps=collect_dependencies(user_object),
        new_jobs=new_jobs,
        deleted_jobs=deleted_jobs,
    )



from typing import List

from .strings_with_escapes import pad_to_screen_length, get_length_on_screen


class TableFormatter(object):
    def __init__(self, sep="|"):
        self.rows = []
        self.cur_row = None
        self.sep = sep

        self.padleft = lambda s, l: pad_to_screen_length(s, l)
        self.strlen = get_length_on_screen

    def row(self):
        if self.cur_row is not None:
            self._push_row()

        self.cur_row = []

    def _push_row(self):
        if self.rows:
            if not len(self.rows[0]) == len(self.cur_row):
                msg = "Invalid row: %s" % str(self.cur_row)
                raise ValueError(msg)
        self.rows.append(self.cur_row)

    def cell(self, s):
        if self.cur_row is None:
            raise ValueError("Call row() before cell().")
        self.cur_row.append(str(s))

    def done(self):
        if self.cur_row is None:
            raise ValueError("Call row() before done().")
        self._push_row()

    def _get_cols(self):
        ncols = len(self.rows[0])
        cols = [list() for _ in range(ncols)]
        for j in range(ncols):
            for r in self.rows:
                cols[j].append(r[j])
        return cols

    def get_lines(self):
        cols = self._get_cols()
        # width of each cols
        wcol = [max(self.strlen(s) for s in col) for col in cols]
        for r in self.rows:
            r = self._get_row_formatted(r, wcol)
            yield r

    def get_lines_multi(self, linewidth, sep="   "):
        """ Gets lines, perhaps multiples on a line """

        # Get all lines
        lines = self.get_lines()
        # Find maximum length of line
        maxlen = max(map(self.strlen, lines))
        #         print('-' * maxlen)
        ncols = 1
        while True:
            l = ncols * maxlen + (ncols - 1) * self.strlen(sep)
            fits = l < linewidth
            #             print(ncols, '->', l)
            if not fits:
                break
            ncols += 1
        ncols -= 1
        ncols = max(1, ncols)

        for mates in groups_match(self.get_lines(), ncols):
            s = sep.join(mates)
            #             print('mates = %r' % mates)
            # FIXME: this raises an error --- (137, 135)
            # assert self.strlen(s) <= linewidth, (self.strlen(s), linewidth)
            if self.strlen(s) > linewidth:
                # raise ValueError()
                pass
            yield s

    def _get_row_formatted(self, row, wcol):
        ss = []
        for j, cell in enumerate(row):
            if wcol[j] > 0:
                entry = self.padleft(cell, wcol[j])
                ss.append(entry)
        return self.sep.join(ss)


def groups_match(it, groupsize) -> List:
    content = list(it)
    gs = list(groups(content, groupsize))
    ngroups = len(gs)

    nrows = ngroups
    ncols = groupsize
    table = []
    for _ in range(nrows):
        row = []
        for _ in range(ncols):
            row.append(None)
        table.append(row)

    for i, x in enumerate(content):
        row = i % nrows
        col = int((i - row) / nrows)
        assert 0 <= col < ncols
        assert 0 <= row < nrows
        table[row][col] = x

    for row in table:
        row = list(filter(lambda _: _ is not None, row))
        yield row


def groups(it, groupsize):
    cur = []
    for x in it:
        cur.append(x)
        if len(cur) == groupsize:
            yield cur
            cur = []
    if cur:
        yield cur



import os
import sys
from typing import Any, Tuple

from compmake.constants import CompmakeConstants
from compmake.events import publish
from compmake.events.registrar import register_handler, remove_all_handlers
from compmake.exceptions import JobFailed, JobInterrupted
from compmake.jobs.actions import make
from compmake.jobs.result_dict import result_dict_check
from compmake.structures import CMJobID
from compmake.utils import setproctitle
from future.moves.queue import Full
from zuper_commons.fs import mkdirs_thread_safe
from zuper_commons.types import check_isinstance

#
# method = multiprocessing.get_start_method(allow_none=True)
# if method is not None:
#     if method !='fork':
#         msg = f'Need "fork", already set to {method}'
#         logger.error(msg)
#     else:
#         multiprocessing.set_start_method("fork")


__all__ = [
    "parmake_job2",
]


def parmake_job2(args: Tuple[CMJobID, Any, str, bool, str]):
    """
    args = tuple job_id, context, queue_name, show_events

    Returns a dictionary with fields "user_object", "new_jobs", 'delete_jobs'.
    "user_object" is set to None because we do not want to
    load in our thread if not necessary. Sometimes it is necessary
    because it might contain a Promise.

    """
    job_id, context, event_queue_name, show_output, logdir = args

    mkdirs_thread_safe(logdir)
    stdout_fn = os.path.join(logdir, f"{job_id}.stdout.log")
    stderr_fn = os.path.join(logdir, f"{job_id}.stderr.log")

    sys.stdout = open(stdout_fn, "w")
    sys.stderr = open(stderr_fn, "w")

    check_isinstance(job_id, str)
    check_isinstance(event_queue_name, str)
    from .pmake_manager import PmakeManager

    # logger.info(f"queues: {PmakeManager.queues}")
    event_queue = PmakeManager.queues[event_queue_name]

    db = context.get_compmake_db()

    setproctitle(f"compmake:{job_id}")

    class G:
        nlostmessages = 0

    try:
        # We register a handler for the events to be passed back
        # to the main process
        def handler(event):
            try:
                if not CompmakeConstants.disable_interproc_queue:
                    event_queue.put(event, block=False)
            except Full:
                G.nlostmessages += 1
                # Do not write messages here, it might create a recursive
                # problem.
                # sys.stderr.write('job %s: Queue is full, message is lost.\n'
                # % job_id)

        remove_all_handlers()

        if show_output:
            register_handler("*", handler)

        def proctitle(event):
            stat = f"[{event.progress}/{event.goal} {event.job_id}] (compmake)"
            setproctitle(stat)

        register_handler("job-progress", proctitle)

        publish(context, "worker-status", job_id=job_id, status="started")

        # Note that this function is called after the fork.
        # All data is conserved, but resources need to be reopened
        try:
            db.reopen_after_fork()
        except:
            pass

        publish(context, "worker-status", job_id=job_id, status="connected")

        res = make(job_id, context=context)

        publish(context, "worker-status", job_id=job_id, status="ended")

        res["user_object"] = None
        result_dict_check(res)
        return res

    except KeyboardInterrupt:
        assert False, "KeyboardInterrupt should be captured by make() (" "inside Job.compute())"
    except JobInterrupted:
        publish(context, "worker-status", job_id=job_id, status="interrupted")
        raise
    except JobFailed:
        raise
    except BaseException:
        # XXX
        raise
    except:
        raise
    finally:
        publish(context, "worker-status", job_id=job_id, status="cleanup")
        setproctitle("compmake-worker-finished %s" % job_id)



import sys

__all__ = [
    "MemoryCache",
]


# noinspection PyArgumentList
class MemoryCache(object):
    def __init__(self, db, cache_values=True):
        self.data = {}
        self.db = db
        self.cache_values = cache_values
        self.keys_to_cache = ":job:"

    def __getitem__(self, key):
        if key in self.data:
            return self.data[key]
        ob = self.db.__getitem__(key)
        if self.should_cache(key):
            self.data[key] = ob
        return ob

    def should_cache(self, key):
        if not self.cache_values:
            return False
        return self.keys_to_cache in key

    def __setitem__(self, key, value):
        # TODO: check value
        if self.should_cache(key):
            self.data[key] = value
        self.db[key] = value

    def __delitem__(self, key):
        if key in self.data:
            self.data.__delitem__(key)
        self.db.__delitem__(key)

    def __contains__(self, key):
        if key in self.data:
            return True
        else:
            return self.db.__contains__(key)

    def sizeof(self, key):
        # XXX: not recursive
        return sys.getsizeof(key)

    def keys(self):
        return self.db.keys()



from .compmake_test import CompmakeTest
from nose.tools import istest


def cases():
    """ Note this uses TestDynamic1.howmany """
    howmany = TestDynamic1.howmany
    assert isinstance(howmany, int)
    print(f"returned {howmany} cases")
    return range(howmany)


def actual_tst(value):
    print("actual_tst(value)")
    return value * 2


def generate_tsts(context, values):
    res = []
    for i, v in enumerate(values):
        res.append(context.comp(actual_tst, v, job_id="actual%d" % i))
    return context.comp(finish, res)


def finish(values):
    return sum(values)


def mockup_dynamic1(context):
    values = context.comp(cases, job_id="values")
    context.comp_dynamic(generate_tsts, values, job_id="generate")


@istest
class TestDynamic1(CompmakeTest):

    howmany = None  # used by cases()

    def test_dynamic1_cleaning(self):
        mockup_dynamic1(self.cc)
        # At this point we have generated only two jobs
        self.assertJobsEqual("all", ["generate", "values"])

        # now we make them
        TestDynamic1.howmany = 3
        self.assert_cmd_success("ls")
        self.assert_cmd_success("make")
        self.assert_cmd_success("ls")

        # this will have created new jobs
        self.assertJobsEqual(
            "all", ["generate", "values", "actual0", "actual1", "actual2", "generate-finish"]
        )
        # ... still to do
        self.assertJobsEqual("todo", ["actual0", "actual1", "actual2", "generate-finish"])

        # we can make them
        self.assert_cmd_success("make")
        self.assert_cmd_success("ls")
        self.assertJobsEqual(
            "done", ["generate", "values", "actual0", "actual1", "actual2", "generate-finish"]
        )

        # Now let's suppose we re-run values and it generates different number of mcdp_lang_tests

        # Now let's increase it to 4
        TestDynamic1.howmany = 4

        self.assert_cmd_success("clean values; make generate")
        self.assert_cmd_success("ls reason=1")

        self.assertJobsEqual(
            "all", ["generate", "values", "actual0", "actual1", "actual2", "actual3", "generate-finish"]
        )
        # some are done
        self.assertJobsEqual(
            "done", ["generate", "values", "actual0", "actual1", "actual2", "generate-finish"]
        )
        # but finish is not updtodate
        self.assertJobsEqual("uptodate", ["generate", "values", "actual0", "actual1", "actual2"])
        # some others are not
        self.assertJobsEqual("todo", ["actual3"])

        # now 2 jobs
        TestDynamic1.howmany = 2
        self.assert_cmd_success("clean values")
        self.assert_cmd_success("ls")
        self.assert_cmd_success("make generate")
        self.assert_cmd_success("ls")

        # Now we should have on job less because actual2 and 3 was not re-defined
        self.assertJobsEqual("all", ["generate", "values", "actual0", "actual1", "generate-finish"])
        # they should be all done, by the way
        self.assertJobsEqual("done", ["generate", "values", "actual0", "actual1", "generate-finish"])



from ..jobs import init_progress_tracking
from compmake import progress

# from contracts import ContractNotRespected
from nose.tools import istest, nottest
import unittest


@istest
class TestProgress(unittest.TestCase):
    def stack_update(self, stack):
        # print "found %s" % stack
        self.stack = stack

    def assert_stack_len(self, d):
        self.assertEqual(d, len(self.stack))

    def setUp(self):
        init_progress_tracking(self.stack_update)

    def test_bad(self):
        """ Many ways to call it in the wrong way. """
        self.assertRaises((ValueError,), progress, "task", 1)

    @nottest  # FIXME, known failure
    def test_hierarchy_flat(self):
        """ Testing basic case. """
        init_progress_tracking(lambda _: None)
        self.assert_stack_len(0)
        progress("A", (0, 2))
        self.assert_stack_len(1)
        progress("A", (1, 2))
        self.assert_stack_len(1)

    @nottest  # FIXME, known failure
    def test_hierarchy_flat2(self):
        data = {}

        def mystack(x):
            data["stack"] = x

        init_progress_tracking(mystack)
        self.assert_stack_len(0)
        progress("A", (0, 2))
        self.assert_stack_len(1)
        progress("B", (0, 2))
        self.assert_stack_len(2)
        progress("B", (1, 2))
        self.assert_stack_len(2)
        progress("A", (1, 2))
        self.assert_stack_len(1)



""" Implements the initial and final banner """
from .. import version
from ..events import register_handler
from ..jobs import all_jobs
from ..ui import compmake_colored
from ..utils import pad_to_screen

compmake_issues_url = "http://github.com/AndreaCensi/compmake/issues"
name = "Compmake"


# banners = [
#     "Tame your Python computations!",
#     "Keep calm and carry on",
# ]


def console_starting(context):
    db = context.get_compmake_db()

    # starting console
    def printb(s):
        print(pad_to_screen(s))

    #     random_banner = random.choice(banners)
    #     banner = "   ``%s,," % random_banner
    #     # banner_s = compmake_colored(banner, 'cyan')

    version_string = "%s %s" % (
        compmake_colored(name, attrs=["bold"]),
        compmake_colored(version, attrs=["bold"]),
    )

    # printb("Welcome to the Compmake console. ")
    njobs = len(list(all_jobs(db)))
    # printb(version_string + ("  (%d jobs loaded)" % njobs) + banner_s)
    printb(version_string + ("  (%d jobs loaded)" % njobs))


def console_ending(event, context):

    url = compmake_colored(compmake_issues_url, attrs=["bold"])
    msg = f"Thanks for using Compmake. Please report problems to {url}"

    print(msg)  # keep print


register_handler("console-starting", console_starting)
register_handler("console-ending", console_ending)



from .memoize_imp import memoized_reset

__all__ = [
    "get_screen_columns",
    "getTerminalSize",
]

import os


@memoized_reset
def get_screen_columns():
    ts = getTerminalSize()
    max_x, _ = ts

    fallback = 90
    if max_x <= 10 or max_x > 1024:
        #         msg = 'Very weird max screen size: %d' % max_x
        #         msg += '\n I will use %s.' % fallback
        #         sys.stderr.write(msg+'\n')

        return fallback
    #         raise ValueError(msg)

    return max_x


def getTerminalSize():
    """
        columns, lines = getTerminalSize()
    """
    cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
    if not cr:
        try:
            fd = os.open(os.ctermid(), os.O_RDONLY)
            cr = ioctl_GWINSZ(fd)
            # noinspection PyTypeChecker
            os.close(fd)
        except:
            cr = (25, 80)

    columns = int(cr[1])
    lines = int(cr[0])

    env = os.environ
    if "COLUMNS" in env:
        columns = int(env["COLUMNS"])
    if "LINES" in env:
        lines = int(env["LINES"])

    return columns, lines


def ioctl_GWINSZ(fd):
    try:
        import fcntl
        import termios
        import struct

        # noinspection PyTypeChecker
        s = fcntl.ioctl(fd, termios.TIOCGWINSZ, "1234")
        return struct.unpack("hh", s)
    except:
        return None



from .compmake_test import CompmakeTest
from nose.tools import istest


def rec(context, n):
    if n == 0:
        return 0
    return context.comp(add, n, context.comp_dynamic(rec, n - 1, job_id="rec-%d" % n), job_id="add-%d" % n)


def add(a, b):
    return a + b


def f(x):
    if not x == 15:
        raise ValueError("Expected 5 + 4 + 3 + 2 + 1 + 0 = 15, not %s" % x)


@istest
class TestDelegation2(CompmakeTest):
    def test_delegation_2(self):
        context = self.cc

        res = context.comp_dynamic(rec, 5, job_id="rec-main")
        context.comp(f, res)

        self.assert_cmd_success("ls")

        self.assert_cmd_success("make")

        self.assert_cmd_success("check-consistency")



from multiprocessing import cpu_count

from .structure import add_config_section, add_config_switch


__all__ = [
    "CONFIG_GENERAL",
    "CONFIG_APPEARANCE",
    "CONFIG_PARALLEL",
]

CONFIG_GENERAL = "General configuration"
CONFIG_APPEARANCE = "Visualization"
CONFIG_PARALLEL = "Multiprocessing backend"
CONFIG_MULTYVAC = "Multyvac backend"

add_config_section(name=CONFIG_GENERAL, desc="", order=-1)
add_config_section(name=CONFIG_APPEARANCE, desc="", order=2)
add_config_section(name=CONFIG_PARALLEL, desc="", order=3)
add_config_section(name=CONFIG_MULTYVAC, desc="", order=4)

add_config_switch(
    "recurse",
    False,
    desc="Default choice for parmake and make whether to run " "generated jobs.",
    section=CONFIG_GENERAL,
)

add_config_switch(
    "new_process",
    False,
    desc="Default choice for parmake and make whether to start a new process for each job.",
    section=CONFIG_GENERAL,
)

add_config_switch(
    "check_params",
    False,
    desc="If true, erases the cache if job parameters appear to change.",
    # Very useful but you need to define __eq__() in all the
    # objects you use as parameters.",
    section=CONFIG_GENERAL,
)

add_config_switch(
    "manager_wait",
    0.1,
    desc="Sleep time, in seconds, to wait if no job has finished. ",
    #                   "Low value gives responsiveness but higher CPU usage",
    section=CONFIG_GENERAL,
)

add_config_switch(
    "echo",
    True,
    desc="Show the output of a job in the console. See " "echo_stdout and echo_stderr.",
    section=CONFIG_APPEARANCE,
)

add_config_switch(
    "echo_stdout",
    True,
    desc="If true and 'echo' is true, the job output to stdout " "is shown.",
    section=CONFIG_APPEARANCE,
)

add_config_switch(
    "echo_stderr",
    True,
    desc="If true and 'echo' is true, the job output to stderr " "is shown.",
    section=CONFIG_APPEARANCE,
)


add_config_switch(
    "details_failed_job",
    True,
    desc="Shows the error details when something fails.",
    section=CONFIG_APPEARANCE,
)

# XXX: same
add_config_switch(
    "status_line_enabled", True, desc="Activate the plugin for status line", section=CONFIG_APPEARANCE
)
add_config_switch(
    "console_status", True, desc="Enables the console_status plugin.", section=CONFIG_APPEARANCE
)
add_config_switch(
    "console_status_style", "normal", desc="status line style (normal, compact)", section=CONFIG_APPEARANCE
)


add_config_switch("colorize", True, desc="Use colors in terminals if possible.", section=CONFIG_APPEARANCE)

add_config_switch(
    "interactive",
    True,
    desc="Assumes that this is an interactive console. (Uses " "\\r to repaint line.)",
    section=CONFIG_APPEARANCE,
)


add_config_switch("console_status_delta", 0.33, desc="Refresh interval (seconds)", section=CONFIG_APPEARANCE)

add_config_switch(
    "readline", True, desc="Try to use readline or pyreadline module.", section=CONFIG_APPEARANCE
)

add_config_switch(
    "set_proc_title",
    True,
    desc="Set the process title to the name of current job.",
    section=CONFIG_APPEARANCE,
)

add_config_switch(
    "verbose_definition",
    False,
    desc="If true, log on stderr about job (re)definition.",
    section=CONFIG_APPEARANCE,
)

add_config_switch(
    "max_parallel_jobs",
    cpu_count(),
    desc="Maximum number of parallel jobs. Default is " "cpu_count().",
    section=CONFIG_PARALLEL,
)

if False:  # To re-implement
    add_config_switch("max_mem_load", 90.0, desc="Maximum physical memory load (%)", section=CONFIG_PARALLEL)

    add_config_switch("max_swap", 20.0, desc="Maximum swap usage (%)", section=CONFIG_PARALLEL)

    add_config_switch(
        "max_cpu_load",
        100.0,
        desc="Maximum CPU load (%). No jobs will be instantiated " "if over threshold.",
        section=CONFIG_PARALLEL,
    )

    add_config_switch(
        "autobal_after",
        cpu_count(),
        # TODO: number of processors / 2
        desc="Autobalances after the given number of processes (%)",
        section=CONFIG_PARALLEL,
    )

    add_config_switch(
        "min_proc_interval",
        0,
        desc="Minimum time interval between instantiating jobs.",
        section=CONFIG_PARALLEL,
    )


add_config_switch(
    "multyvac_debug", False, desc="If true, shows multyvac's logging output.", section=CONFIG_MULTYVAC
)

add_config_switch(
    "multyvac_max_jobs", 50, desc="Default number of cloud jobs to be instantiated", section=CONFIG_MULTYVAC
)

add_config_switch("multyvac_layer", "", desc="Multyvac 'layer'", section=CONFIG_MULTYVAC)

add_config_switch(
    "multyvac_sync_down", "", desc="Multyvac synchronization directory (output)", section=CONFIG_MULTYVAC
)

add_config_switch(
    "multyvac_sync_up", "", desc="Multyvac synchronization directory (input)", section=CONFIG_MULTYVAC
)

add_config_switch("multyvac_core", "c2", desc="Multyvac core (c1,c2,f2)", section=CONFIG_MULTYVAC)



from .visualization import compmake_colored

__all__ = [
    "colorize_loglevel",
]


def colorize_loglevel(levelno, msg):
    # TODO: use Compmake's way
    if levelno >= 50:
        return compmake_colored(msg, "red")
    elif levelno >= 40:
        return compmake_colored(msg, "red")
    elif levelno >= 30:
        return compmake_colored(msg, "yellow")
    elif levelno >= 20:
        return compmake_colored(msg, "green")
    elif levelno >= 10:
        return compmake_colored(msg, "cyan")
    else:
        return msg



from nose.tools import istest

from .compmake_test import CompmakeTest
from .mockup import mockup_recursive_5


@istest
class TestParmakeNewProcess(CompmakeTest):

    # TODO: parmake_pool
    def test_parmake_new_process(self):
        mockup_recursive_5(self.cc)
        self.assert_cmd_success("parmake recurse=1 new_process=1;ls")

    # TODO: parmake_pool
    def test_make_new_process(self):
        mockup_recursive_5(self.cc)
        self.assert_cmd_success("make recurse=1 new_process=1;ls")



from compmake.structures import Cache

__all__ = ["compute_priorities"]


def compute_priorities(all_targets, cq, priorities=None):
    """ Computes the priority for all_targets.

        :param priorities: str->float: cache
    """
    if priorities is None:
        priorities = {}
    all_targets = set(all_targets)
    for job_id in all_targets:
        p = compute_priority(job_id=job_id, priorities=priorities, targets=all_targets, cq=cq)
        priorities[job_id] = p
    return priorities


def compute_priority(job_id, priorities, targets, cq):
    """ Computes the priority for one job. It uses caching results in
        self.priorities if they are found. """
    if job_id in priorities:
        return priorities[job_id]

    parents = set(cq.direct_parents(job_id))
    parents_which_are_targets = [x for x in parents if x in targets]

    # Dynamic jobs get bonus
    job = cq.get_job(job_id)
    if job.needs_context:
        base_priority = 10
    else:
        base_priority = -1

    if not parents:
        # top level target
        base_priority += 5

    cache = cq.get_job_cache(job_id)
    if cache.state == Cache.FAILED:
        base_priority -= 100

    if not parents_which_are_targets:
        priority = base_priority
    else:
        pf = lambda p: compute_priority(p, priorities, targets, cq=cq)
        # it was -1
        parents_priority = list(map(pf, parents_which_are_targets))
        # priority = base_priority + max(parents_priority)
        priority = base_priority + sum(parents_priority)

    priorities[job_id] = priority

    return priority



import re

from ..exceptions import UserError


__all__ = [
    "wildcard_to_regexp",
    "expand_wildcard",
]


def wildcard_to_regexp(arg):
    """ Returns a regular expression from a shell wildcard expression. """
    return re.compile("\A" + arg.replace("*", ".*") + "\Z")


def expand_wildcard(wildcard, universe):
    """ Expands a wildcard expression against the given list.
        wildcard: string with '*'
        universe: list of strings
    """
    assert wildcard.find("*") > -1
    regexp = wildcard_to_regexp(wildcard)
    num_matches = 0
    for x in universe:
        if regexp.match(x):
            num_matches += 1
            yield x
    if num_matches == 0:
        msg = 'Could not find matches for pattern "%s".' % wildcard
        raise UserError(msg)



""" These are the commands available from the CLI.

There are 3 special variables:
- 'args': list of all command line arguments
- 'job_list': the remaining argument parsed as a job list.
- 'non_empty_job_list': same, but error if not specified.
"""
from .console import ask_question
from .helpers import ACTIONS, COMMANDS_ADVANCED, GENERAL, ui_command, ui_section
from .visualization import ui_error, ui_info
from .. import CompmakeConstants, get_compmake_status
from ..exceptions import JobFailed, MakeFailed, ShellExitRequested, UserError
from ..jobs import all_jobs, CacheQueryDB
from ..jobs.actions import clean_targets
from ..utils import safe_pickle_dump

ui_section(GENERAL)

__all__ = [
    "make_single",
    "quit",
    "raise_error_if_manager_failed",
    "ask_if_sure_remake",
]


# noinspection PyUnusedLocal,PyShadowingBuiltins
@ui_command(alias=["exit"])
def quit(context):
    """ Exits Compmake's console. """
    raise ShellExitRequested()


def raise_error_if_manager_failed(manager):
    """
    Raises MakeFailed if there are failed jobs in the manager.

    :param manager: The Manager
    """
    if manager.failed:
        raise MakeFailed(failed=manager.failed, blocked=manager.blocked)


@ui_command(section=COMMANDS_ADVANCED, dbchange=True)
def delete(job_list, context):
    """ Remove completely the job from the DB. Useful for generated jobs (
    "delete not root"). """
    from compmake.jobs.storage import delete_all_job_data

    job_list = [x for x in job_list]

    db = context.get_compmake_db()
    for job_id in job_list:
        delete_all_job_data(job_id=job_id, db=db)


@ui_command(section=ACTIONS, dbchange=True)
def clean(job_list, context, cq: CacheQueryDB):
    """
        Cleans the result of the selected computation (or everything if
        nothing specified).

        If cleaning a dynamic job, it *deletes* all jobs it created.

    """
    db = context.get_compmake_db()

    # job_list = list(job_list) # don't ask me why XXX
    job_list = [x for x in job_list]

    if not job_list:
        job_list = list(all_jobs(db=db))

    if not job_list:
        return

    # Use context
    if get_compmake_status() == CompmakeConstants.compmake_status_interactive:
        question = f"Should I clean {len(job_list)} jobs? [y/n] "
        answer = ask_question(question)
        if not answer:
            ui_info(context, "Not cleaned.")
            return

    # ui_info(context, f'Going to clean {job_list}')
    clean_targets(job_list, db=db, cq=cq)


# TODO: add hidden
@ui_command(section=COMMANDS_ADVANCED, dbchange=True)
def make_single(job_list, context, out_result):
    """ Makes a single job -- not for users, but for slave mode. """
    if len(job_list) > 1:
        raise UserError("I want only one job")

    job_id = job_list[0]

    from compmake.jobs import actions

    try:
        # info('making job %s' % job_id)
        res = actions.make(job_id=job_id, context=context)
        # info('Writing to %r' % out_result)
        safe_pickle_dump(res, out_result)
        return 0
    except JobFailed as e:
        # info('Writing to %r' % out_result)
        safe_pickle_dump(e.get_result_dict(), out_result)
        raise MakeFailed(failed=[job_id])
    except BaseException as e:
        ui_error(context, f"warning: {e}")
        raise


def ask_if_sure_remake(non_empty_job_list):
    """ If interactive, ask the user yes or no. Otherwise returns True. """
    if get_compmake_status() == CompmakeConstants.compmake_status_interactive:
        question = f"Should I clean and remake {len(non_empty_job_list)} jobs? [y/n] "
        answer = ask_question(question)
        if not answer:
            # info("Not cleaned.")
            return False
        else:
            return True
    return True



from setuptools import find_packages, setup


def get_version(filename):
    import ast

    version_ = None
    with open(filename) as f:
        for line in f:
            if line.startswith("__version__"):
                version_ = ast.parse(line).body[0].value.s
                break
        else:
            raise ValueError("No version found in %r." % filename)
    if version_ is None:
        raise ValueError(filename)
    return version_


version = get_version(filename="src/compmake/__init__.py")

line = "z6"

install_requires = [
    "PyContracts3",
    "termcolor",
    "setproctitle",
    "PyYaml",
    "psutil",
    "dill",
    "coverage",
    "decorator",
    "SystemCmd-z6",
    "future",
    "networkx>=2,<3",
    "six",
    "zuper-commons-z6>=6.1.7",
    "asciimatics",
]

setup(
    name=f"compmake-{line}",
    author="Andrea Censi",
    url="http://compmake.org",
    version=version,
    description="Compmake is a non-obtrusive module that provides "
    "'make'-like facilities to your Python applications,"
    "including caching of results, robustness to jobs failure, "
    "and multiprocessing/multihost parallel processing.",
    long_description="""
        Compmake is a non-obtrusive module that provides
        'make'-like facilities to your Python applications,
        including caching of results, robustness to jobs failure,
        and multiprocessing/multihost parallel processing.

        Please see for docs: http://compmake.org
        and get the manual PDF at: http://purl.org/censi/compmake-manual
    """,
    keywords="parallel processing, make, cmake, manager, recovery",
    license="LGPL",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: System :: Clustering",
        "Topic :: System :: Distributed Computing",
        "Topic :: System :: Hardware :: Symmetric Multi-processing",
        "License :: OSI Approved :: GNU Library or " "Lesser General Public License (LGPL)",
    ],
    package_dir={"": "src"},
    packages=find_packages("src"),
    entry_points={
        "console_scripts": [
            "compmake = compmake.scripts.master:main",
            "zuper-make = compmake.make_bridge:make_bridge_main",
            "zmake = compmake.make_bridge:make_bridge_main",
            # 'compmake_slave = compmake.jobs.manager_ssh_cluster:compmake_slave'
        ]
    },
    install_requires=install_requires,
    tests_require=["nose"],
)



from nose.tools import istest

from compmake.unittests.compmake_test import CompmakeTest

from ..exceptions import UserError


def uses_nested(context):
    def f1():
        pass

    context.comp(f1)


def uses_lambda(context):

    context.comp(lambda x: x, 1)


@istest
class TestInvalidFunctions(CompmakeTest):
    def test_invalid_function_nested(self):
        self.assertRaises(UserError, uses_nested, self.cc)

    def test_invalid_function_lambda(self):
        self.assertRaises(UserError, uses_lambda, self.cc)



from compmake.exceptions import UserError
from compmake.utils import which
from zuper_commons.types import raise_wrapped

__all__ = [
    "check_sge_environment",
]


def check_sge_environment():
    msg_install = " Please install SGE properly. "
    try:
        _ = which("qsub")
    except ValueError as e:
        msg = 'Program "qsub" not available.\n'
        msg += msg_install
        raise_wrapped(UserError, e, msg)



import os
import pickle

from zuper_typing import debug_print
from ..jobs import get_job_userobject, get_job_userobject_resolved, is_job_userobject_available
from ..ui import COMMANDS_ADVANCED, ui_command
from ..ui.visualization import ui_info, ui_message


@ui_command(section=COMMANDS_ADVANCED)
def dump(non_empty_job_list, context, directory="."):
    """ Dumps the result of jobs as pickle files.

        Arguments:
            directory='.'   where to dump the files

    """
    db = context.get_compmake_db()
    for job_id in non_empty_job_list:

        if is_job_userobject_available(job_id, db=db):
            user_object = get_job_userobject(job_id, db=db)
            filename = os.path.join(directory, job_id + ".pickle")
            with open(filename, "wb") as f:
                pickle.dump(user_object, f)
            ui_info(context, f"Wrote {filename}")
        else:
            ui_message(context, f"Job {job_id} is not ready yet.")


@ui_command(section=COMMANDS_ADVANCED)
def dump_stdout(non_empty_job_list, context, resolve=False):
    """ Dumps the result of jobs on stdout. """
    db = context.get_compmake_db()
    for job_id in non_empty_job_list:
        if is_job_userobject_available(job_id, db=db):
            if resolve:
                user_object = get_job_userobject_resolved(job_id, db)
            else:
                user_object = get_job_userobject(job_id, db=db)
            ui_message(context, debug_print(user_object))
        else:
            ui_message(context, f"Job {job_id} is not ready yet.")



# from contracts import contract
import os, sys
from contextlib import contextmanager


# @contract(returns=dict)
from typing import Dict


def pickle_main_context_save() -> Dict:
    """ Remember who was the __main__ module """
    module = sys.modules["__main__"]
    filename = module.__file__
    name = os.path.splitext(os.path.basename(filename))[0]
    main_module = name
    main_path = os.path.realpath(os.path.dirname(filename))
    return dict(main_module=main_module, main_path=main_path)


@contextmanager
def pickle_main_context_load(c):
    main_path = c["main_path"]
    main_module = c["main_module"]

    try:
        if not main_path in sys.path:
            sys.path.append(main_path)

        cur_main = sys.modules["__main__"]

        try:
            m = __import__(main_module, fromlist=["dummy"])
            m.__name__ = "__main__"
            sys.modules["__main__"] = m
        except ImportError as e:
            #             print('pickle_main_context_load: Cannot import %r: %s'
            #                   % (main_module, e))
            pass
        yield

    finally:
        sys.modules["__main__"] = cur_main



from nose.tools import istest
from .compmake_test import CompmakeTest
from .mockup import mockup2_nofail


@istest
class PluginsTest(CompmakeTest):
    def mySetUp(self):
        mockup2_nofail(self.cc)

    def testDetails(self):
        jobs = self.get_jobs("all")
        for job_id in jobs:
            self.assert_cmd_success("details %s" % job_id)
        self.assert_cmd_success("details %s %s" % (jobs[0], jobs[1]))

    def testList(self):
        jobs = self.get_jobs("all")
        self.assert_cmd_success("ls")
        self.assert_cmd_success("ls %s" % jobs[0])

        # empty list
        # self.assert_cmd_success('ls block* and done')

    # def testCredits(self):
    #     self.assert_cmd_success('credits')

    def testCheckConsistency(self):
        self.assert_cmd_success("check-consistency")

    def testDump(self):
        dirname = self.cc.get_compmake_db().basepath
        jobs = self.get_jobs("done")
        for job_id in jobs:
            self.assert_cmd_success("dump directory=%s %s" % (dirname, job_id))

        # TODO: add check that it fails for not done
        jobs = self.get_jobs("not done")
        for job_id in jobs:
            self.assert_cmd_success("dump directory=%s %s" % (dirname, job_id))



from .. import CompmakeGlobalState

__all__ = [
    "create_config_html",
]


def create_config_html(file):  # @ReservedAssignment
    config_switches = CompmakeGlobalState.config_switches
    config_sections = CompmakeGlobalState.config_sections

    # TODO: HTML escaping?
    ordered_sections = sorted(config_sections.values(), key=lambda _section: _section.order)

    file.write("<table class='compmake-config'>\n")

    for section in ordered_sections:
        file.write("<tr><th colspan='3'>%s</th></tr/ \n" % section.name)

        if section.desc:
            file.write("<tr><td colspan='3'> %s </td></tr> \n" % section.desc)

        for name in section.switches:
            switch = config_switches[name]
            desc = str(switch.desc)

            file.write(
                "<tr> <td class='config-name'><tt>%s</tt></td> \
<td class='config-value'><tt>%s<tt></td> \
<td class='config-desc'>%s</td> </tr> \n"
                % (name, switch.default_value, desc)
            )

    file.write("</table>\n")

    file.write(
        """
    <style type="text/css">
    .config-value: { padding: 1em }

    </style>

    """
    )



from compmake.events import register_handler
from compmake.state import get_compmake_config
from compmake.ui.visualization import ui_error, ui_info
from zuper_commons.text import indent


# TODO: command-succeeded: {'command': '
# command-interrupted: {'reason': 'KeyboardInterrupt', 'command': 'ls todo'}
def command_interrupted(context, event):
    c = event.kwargs["command"]
    traceback = event.kwargs["traceback"]
    ui_error(context, f"Command {c!r} interrupted.")
    ui_error(context, traceback)


register_handler("command-interrupted", command_interrupted)


def command_failed(context, event):
    c = event.kwargs["command"]
    r = event.kwargs["reason"]
    ui_error(context, f"Command {c!r} failed: {r}")


register_handler("command-failed", command_failed)

# my_prefix = '(plugin commands_status) '
my_prefix = ""


def command_line_interrupted(context, event):
    # Only write something if it is more than one
    command = event.kwargs["command"]
    if not ";" in command:
        return
    ui_error(context, my_prefix + f"Command sequence {command!r} interrupted.")


register_handler("command-line-interrupted", command_line_interrupted)


def command_line_failed(context, event):
    # Only write something if it is more than one
    command = event.kwargs["command"]
    if not ";" in command:
        return
    ui_error(context, my_prefix + f"Command sequence {command!r} failed.")


register_handler("command-line-failed", command_line_failed)


def job_failed(context, event):
    yes = get_compmake_config("details_failed_job")
    if not yes:
        return
    job_id = event.kwargs["job_id"]
    reason = event.kwargs["reason"]
    bt = event.kwargs["bt"]

    msg = f"Job {job_id!r} failed:"
    # s = reason.strip
    # if get_compmake_config('echo'):
    #     s += '\n' + bt
    msg += "\n" + indent(reason.strip(), "| ")

    # if get_compmake_config("echo"):
    #     s = bt.strip()
    #     msg += "\n" + indent(s, "> ")
    # else:
    #     msg += '\nUse "config echo 1" to have errors displayed.'
    msg += f'\nWrite "details {job_id}" to inspect the error.'
    ui_error(context, my_prefix + msg)


register_handler("job-failed", job_failed)


def job_interrupted(context, event):
    ui_error(
        context,
        my_prefix + "Job %r interrupted:\n %s" % (event.kwargs["job_id"], indent(event.kwargs["bt"], "> ")),
    )


register_handler("job-interrupted", job_interrupted)


def compmake_bug(context, event):
    ui_error(context, my_prefix + event.kwargs["user_msg"])
    ui_error(context, my_prefix + event.kwargs["dev_msg"])


register_handler("compmake-bug", compmake_bug)


# We ignore some other events; otherwise they will be catched
# by the default handler
def ignore(context, event):
    pass


register_handler("command-starting", ignore)
register_handler("command-line-starting", ignore)
register_handler("command-line-failed", ignore)
register_handler("command-line-succeeded", ignore)
register_handler("command-line-interrupted", ignore)

register_handler("manager-phase", ignore)

register_handler("parmake-status", ignore)

register_handler("job-succeeded", ignore)
register_handler("job-interrupted", ignore)

if True:  # debugging
    register_handler("worker-status", ignore)
    register_handler("manager-job-done", ignore)
    register_handler("manager-job-failed", ignore)
    register_handler("manager-job-processing", ignore)


def manager_succeeded(context, event):
    if event.kwargs["nothing_to_do"]:
        ui_info(context, "Nothing to do.")
    else:
        ntargets = len(event.kwargs["all_targets"])
        ndone = len(event.kwargs["done"])
        nfailed = len(event.kwargs["failed"])
        nblocked = len(event.kwargs["blocked"])
        if ntargets:
            s = f"Processed {ntargets} jobs ("
            ss = []
            if ndone:
                ss.append(f"{ndone} done")
            if nfailed:
                ss.append(f"{nfailed} failed")
            if nblocked:
                ss.append(f"{nblocked} blocked")
            s += ", ".join(ss)
            s += ")."
            if nfailed:
                ui_error(context, s)
            else:
                ui_info(context, s)


register_handler("manager-succeeded", manager_succeeded)  # TODO: maybe write sth



import inspect
import traceback
from typing import cast, Dict, List, Set

from zuper_commons.types import check_isinstance, describe_type, describe_value, raise_wrapped
from .helpers import get_commands, UIState
from .visualization import ui_warning
from .. import CompmakeConstants, get_compmake_config, get_compmake_status
from ..constants import DefaultsToConfig
from ..context import Context
from ..events import publish
from ..exceptions import CommandFailed, CompmakeBug, UserError
from ..jobs import (
    all_jobs,
    CacheQueryDB,
    collect_dependencies,
    get_job,
    job_exists,
    parse_job_list,
    set_job,
    set_job_args,
)
from ..jobs.actions import clean_cache_relations
from ..jobs.storage import db_job_add_parent_relation, get_job_args
from ..structures import CMJobID, Job, Promise, same_computation
from ..utils import interpret_strings_like, try_pickling


def generate_job_id(base, context):
    """
        Generates a unique job_id for the specified commmand.
        Takes into account job_prefix if that's defined.
    """

    stack = context.currently_executing
    # print('generating an ID with base = %s and stack %s' % (base, stack))
    job_prefix = context.get_comp_prefix()
    # Use the job id as prefix
    if job_prefix is None and len(stack) > 1:
        job_prefix = stack[-1]

    max_options = 1000 * 1000

    def get_options():
        counters = context.generate_job_id_counters
        if not job_prefix in counters:
            counters[job_prefix] = 2

        if job_prefix:
            yield "%s-%s" % (job_prefix, base)
            while counters[job_prefix] <= max_options:
                yield "%s-%s-%d" % (job_prefix, base, counters[job_prefix])
                counters[job_prefix] += 1
        else:
            yield base
            while counters[job_prefix] <= max_options:
                yield "%s-%d" % (base, counters[job_prefix])
                counters[job_prefix] += 1

    db = context.get_compmake_db()
    cq = CacheQueryDB(db)
    for x in get_options():
        check_isinstance(x, str)
        defined = context.was_job_defined_in_this_session(x)
        if defined:
            continue
        exists = defined or cq.job_exists(x)
        if not exists:
            # print('u')
            return x
        else:
            # if it is the same job defined in the same stack
            defined_by = cq.get_job(x).defined_by
            # print('a')
            # print('  Found, he was defined by %s' % defined_by)
            if defined_by == stack:
                # print('x')
                return x
            else:
                # print('-')
                continue

    raise CompmakeBug("Could not generate a job id")


def clean_other_jobs(context):
    """ Cleans jobs not defined in the session """
    # print('cleaning other jobs. Defined: %r' %
    # context.get_jobs_defined_in_this_session())
    db = context.get_compmake_db()
    if get_compmake_status() == CompmakeConstants.compmake_status_slave:
        return
    #     from .console import ask_question
    #
    #     answers = {'a': 'a', 'n': 'n', 'y': 'y', 'N': 'N'}
    #
    #     if is_interactive_session():
    #         clean_all = False
    #     else:
    #         clean_all = True

    # logger.info('Cleaning all jobs not defined in this session.'
    #                ' Previous: %d' % len(defined_now))

    todelete = set()

    for job_id in all_jobs(force_db=True, db=db):
        if not context.was_job_defined_in_this_session(job_id):
            # it might be ok if it was not defined by ['root']
            job = get_job(job_id, db=db)
            if job.defined_by != ["root"]:
                # keeping this around
                continue

            who = job.defined_by[1:]
            if who:
                a = "->".join(who)
                defined = f" (defined by {a})"
            else:
                defined = ""

            from compmake.ui.visualization import ui_info

            ui_info(context, f"Job {job_id!r} not defined in this session {defined}; cleaning.")
            #
            #             if not clean_all:
            #                 # info('Job %s defined-by %s' % (job_id, job.defined_by))
            #                 text = ('Found spurious job %s; cleaning? '
            #                         '[y]es, [a]ll, [n]o, [N]one ' % job_id)
            #                 answer = ask_question(text, allowed=answers)
            #
            #                 if answer == 'n':
            #                     continue
            #
            #                 if answer == 'N':
            #                     break
            #
            #                 if answer == 'a':
            #                     clean_all = True
            #             else:
            #                 pass
            #                 #logger.info('Cleaning job: %r (defined by %s)' % (job_id,
            #                 # job.defined_by))

            todelete.add(job_id)
    delete_jobs_recurse_definition(todelete, db)


# @contract(returns="set(unicode)")
def delete_jobs_recurse_definition(jobs, db) -> Set[str]:
    """ Deletes all jobs given and the jobs that they defined.
        Returns the set of jobs deleted. """
    from compmake.jobs.queries import definition_closure

    closure = definition_closure(jobs, db)

    all_jobs = jobs | closure
    for job_id in all_jobs:
        clean_cache_relations(job_id, db)

    for job_id in all_jobs:
        from compmake.jobs.storage import delete_all_job_data

        delete_all_job_data(job_id, db)

    return all_jobs


class WarningStorage(object):
    warned = set()


def comp_(context, command_, *args, **kwargs):
    """
        Main method to define a computation step.

        Extra arguments:

        :arg:job_id:   sets the job id (respects job_prefix)
        :arg:extra_dep: extra dependencies (not passed as arguments)
        :arg:command_name: used to define job name if job_id not provided.
        If not given, command_.__name__ is used.

        :arg:needs_context: if this is a dynamic job

        Raises UserError if command is not pickable.
    """

    db = context.get_compmake_db()

    command = command_

    if hasattr(command, "__module__") and command.__module__ == "__main__":
        if not command in WarningStorage.warned:
            if WarningStorage.warned:
                # already warned for another function
                msg = "(Same warning for function %r.)" % command.__name__
            else:
                msg = "A warning about the function %r: " % command.__name__
                msg += (
                    "This function is defined directly in the __main__ "
                    "module, "
                    "which means that it cannot be pickled correctly due to "
                    "a limitation of Python and 'make new_process=1' will "
                    "fail. "
                    "For best results, please define functions in external "
                    "modules. "
                    "For more info, read "
                    "http://stefaanlippens.net/pickleproblem "
                    "and the bug report http://bugs.python.org/issue5509."
                )
            ui_warning(context, msg)
            WarningStorage.warned.add(command)

    if get_compmake_status() == CompmakeConstants.compmake_status_slave:
        return None

    # Check that this is a pickable function
    try:
        try_pickling(command)
    except Exception as e:
        msg = (
            "Cannot pickle function. Make sure it is not a lambda "
            "function or a nested function. (This is a limitation of "
            "Python)"
        )
        raise_wrapped(UserError, e, msg, command=command)

    if CompmakeConstants.command_name_key in kwargs:
        command_desc = kwargs.pop(CompmakeConstants.command_name_key)
    elif hasattr(command, "__name__"):
        command_desc = command.__name__

    else:
        command_desc = type(command).__name__

    args = list(args)  # args is a non iterable tuple

    # Get job id from arguments
    if CompmakeConstants.job_id_key in kwargs:
        # make sure that command does not have itself a job_id key
        try:
            argspec = inspect.getfullargspec(command)
        except TypeError:
            # Assume Cython function
            # XXX: write test
            pass
        else:
            if CompmakeConstants.job_id_key in argspec.args:
                msg = (
                    "You cannot define the job id in this way because %r "
                    "is already a parameter of this function." % CompmakeConstants.job_id_key
                )
                raise UserError(msg)

        job_id = kwargs[CompmakeConstants.job_id_key]
        check_isinstance(job_id, str)
        if " " in job_id:
            msg = "Invalid job id: %r" % job_id
            raise UserError(msg)

        job_prefix = context.get_comp_prefix()
        if job_prefix:
            job_id = "%s-%s" % (job_prefix, job_id)

        del kwargs[CompmakeConstants.job_id_key]

        if context.was_job_defined_in_this_session(job_id):
            # unless it is dynamically geneterated
            if not job_exists(job_id, db=db):
                msg = "The job %r was defined but not found in DB. I will let it slide." % job_id
                print(msg)
            else:
                msg = "The job %r was already defined in this session." % job_id
                old_job = get_job(job_id, db=db)
                msg += "\n  old_job.defined_by: %s " % old_job.defined_by
                msg += "\n context.currently_executing: %s " % context.currently_executing
                msg += " others defined in session: %s" % context.get_jobs_defined_in_this_session()
                print(msg)
                #                 warnings.warn('I know something is more complicated here')
                #             if old_job.defined_by is not None and
                # old_job.defined_by == context.currently_executing:
                #                 # exception, it's ok
                #                 pass
                #             else:

                msg = "Job %r already defined." % job_id
                raise UserError(msg)
        else:
            if job_exists(job_id, db=db):
                # ok, you gave us a job_id, but we still need to check whether
                # it is the same job
                stack = context.currently_executing
                defined_by = get_job(job_id, db=db).defined_by
                if defined_by == stack:
                    # this is the same job-redefining
                    pass
                else:

                    for i in range(1000):  # XXX
                        n = cast(CMJobID, f"{job_id}-{i}")
                        if not job_exists(n, db=db):
                            job_id = n
                            break

                    if False:
                        print("The job_id %r was given explicitly but already " "defined." % job_id)
                        print("current stack: %s" % stack)
                        print("    its stack: %s" % defined_by)
                        print("New job_id is %s" % job_id)

    else:
        job_id = generate_job_id(command_desc, context=context)

    context.add_job_defined_in_this_session(job_id)

    # could be done better
    if "needs_context" in kwargs:
        needs_context = True
        del kwargs["needs_context"]
    else:
        needs_context = False

    if CompmakeConstants.extra_dep_key in kwargs:
        extra_dep = kwargs[CompmakeConstants.extra_dep_key]
        del kwargs[CompmakeConstants.extra_dep_key]

        if not isinstance(extra_dep, (list, Promise)):
            msg = 'The "extra_dep" argument must be a list of promises; ' "got: %s" % describe_value(
                extra_dep
            )
            raise ValueError(msg)
        if isinstance(extra_dep, Promise):
            extra_dep = [extra_dep]
        assert isinstance(extra_dep, list)
        for ed in extra_dep:
            if not isinstance(ed, Promise):
                msg = 'The "extra_dep" argument must be a list of promises; ' "got: %s" % describe_value(
                    extra_dep
                )
                raise ValueError(msg)
        extra_dep = collect_dependencies(extra_dep)

    else:
        extra_dep = set()

    children = collect_dependencies([args, kwargs])
    children.update(extra_dep)

    for c in children:
        if not job_exists(c, db):
            msg = "Job %r references a job %r that doesnt exist." % (job_id, c)
            raise ValueError(msg)

    all_args = (command, args, kwargs)

    assert len(context.currently_executing) >= 1
    assert context.currently_executing[0] == "root"

    c = Job(
        job_id=job_id,
        children=children,
        command_desc=command_desc,
        needs_context=needs_context,
        defined_by=context.currently_executing,
    )

    # Need to inherit the pickle
    if context.currently_executing[-1] != "root":
        parent_job = get_job(context.currently_executing[-1], db)
        c.pickle_main_context = parent_job.pickle_main_context

    if job_exists(job_id, db):
        old_job = get_job(job_id, db)

        if old_job.defined_by != c.defined_by:
            warning("Redefinition of %s: " % job_id)
            warning(" cur defined_by: %s" % c.defined_by)
            warning(" old defined_by: %s" % old_job.defined_by)

        if old_job.children != c.children:
            # warning('Redefinition problem:')
            # warning(' old children: %s' % (old_job.children))
            # warning(' old dyn children: %s' % old_job.dynamic_children)
            # warning(' new children: %s' % (c.children))

            # fixing this
            for x, deps in old_job.dynamic_children.items():
                if not x in c.children:
                    # not a child any more
                    # FIXME: ok but note it might be a dependence of a child
                    # continue
                    pass
                c.dynamic_children[x] = deps
                for j in deps:
                    if not j in c.children:
                        c.children.add(j)

        if old_job.parents != c.parents:
            # warning('Redefinition of %s: ' % job_id)
            #  warning(' cur parents: %s' % (c.parents))
            # warning(' old parents: %s' % old_job.parents)
            for p in old_job.parents:
                c.parents.add(p)

                # TODO: preserve defines
                #     from compmake.ui.visualization import info
                #     info('defining job %r with children %r' % (job_id,
                # c.children))

                #     if True or c.defined_by == ['root']:

    for child in children:
        db_job_add_parent_relation(child=child, parent=job_id, db=db)

    if get_compmake_config("check_params") and job_exists(job_id, db):
        # OK, this is going to be black magic.
        # We want to load the previous job definition,
        # however, by unpickling(), it will start
        # __import__()ing the modules, perhaps
        # even the one that is calling us.
        # What happens, then is that it will try to
        # add another time this computation recursively.
        # What we do, is that we temporarely switch to
        # slave mode, so that recursive calls to comp()
        # are disabled.
        #             old_status = get_compmake_status()
        #             set_compmake_status(
        # CompmakeConstants.compmake_status_slave)
        all_args_old = get_job_args(job_id, db=db)
        #             set_compmake_status(old_status)
        same, reason = same_computation(all_args, all_args_old)

        if not same:
            # print('different job, cleaning cache:\n%s  ' % reason)
            from compmake.jobs.actions import clean_targets

            clean_targets([job_id], db)
            #             if job_cache_exists(job_id, db):
            #                 delete_job_cache(job_id, db)
            publish(context, "job-redefined", job_id=job_id, reason=reason)
        else:
            # print('ok, same job')
            pass
            # XXX TODO clean the cache
            #             else:
            #                 publish(context, 'job-already-defined',
            # job_id=job_id)

    set_job_args(job_id, all_args, db=db)
    set_job(job_id, c, db=db)
    publish(context, "job-defined", job_id=job_id)

    return Promise(job_id)


# @contract(commands_str="unicode", context=Context, cq=CacheQueryDB, returns="None")
def interpret_commands(commands_str: str, context: Context, cq: CacheQueryDB, separator=";") -> None:
    """
        Interprets what could possibly be a list of commands (separated by ";")

        Returns None
    """
    if not isinstance(commands_str, str):
        msg = "I expected a string, got %s." % describe_type(commands_str)
        raise ValueError(msg)

    # split with separator
    commands = commands_str.split(separator)
    # remove extra spaces
    commands = [x.strip() for x in commands]
    # filter dummy commands
    commands = [x for x in commands if x]

    if not commands:
        # nothing to do
        return None

    for cmd in commands:
        try:
            publish(context, "command-starting", command=cmd)
            # noinspection PyNoneFunctionAssignment
            retcode = interpret_single_command(cmd, context=context, cq=cq)

        except KeyboardInterrupt:
            publish(
                context,
                "command-interrupted",
                command=cmd,
                reason="KeyboardInterrupt",
                traceback=traceback.format_exc(),
            )
            raise
        except UserError as e:
            publish(context, "command-failed", command=cmd, reason=e)
            raise
        # TODO: all the rest is unexpected

        if retcode == 0 or retcode is None:
            continue
        else:
            if isinstance(retcode, int):
                publish(context, "command-failed", command=cmd, reason=f"Return code {retcode!r}")
                raise CommandFailed(f"ret code {retcode}")
            else:
                publish(context, "command-failed", command=cmd, reason=retcode)
                raise CommandFailed(f"ret code {retcode}")


# @contract(returns="None", commands_line="unicode")
def interpret_single_command(commands_line: str, context, cq: CacheQueryDB):
    """ Returns None or raises CommandFailed """
    if not isinstance(commands_line, str):
        raise ValueError("Expected a string")

    ui_commands = get_commands()

    commands = commands_line.split()

    command_name = commands[0]

    # Check if this is an alias
    if command_name in UIState.alias2name:
        command_name = UIState.alias2name[command_name]

    if not command_name in ui_commands:
        msg = "Unknown command %r (try 'help'). " % command_name
        raise UserError(msg)

    # XXX: use more elegant method
    cmd = ui_commands[command_name]
    dbchange = cmd.dbchange
    function = cmd.function

    args = commands[1:]

    # look for  key=value pairs
    other = []
    kwargs = {}

    signature = inspect.signature(function)

    defaults = get_defaults(signature)
    args_without_default = get_args_without_defaults(signature)

    for a in args:
        if a.find("=") > 0:
            k, v = a.split("=")

            if not k in signature.parameters:
                msg = (
                    f"You passed the argument {k!r} for command {cmd.name!r}, "
                    f"but the only available arguments are {signature.parameters}."
                )
                raise UserError(msg)

            # look if we have a default value
            if not k in defaults:
                # no default, pass as string
                kwargs[k] = v
            else:
                default_value = defaults[k]

                if isinstance(default_value, DefaultsToConfig):
                    default_value = get_compmake_config(default_value.switch)
                try:
                    kwargs[k] = interpret_strings_like(v, default_value)
                except ValueError:
                    msg = "Could not parse %s=%s as %s." % (k, v, type(default_value))
                    raise UserError(msg)
        else:
            other.append(a)

    args = other

    function_args = signature.parameters
    # set default values
    for argname, argdefault in defaults.items():
        if not argname in kwargs and isinstance(argdefault, DefaultsToConfig):
            v = get_compmake_config(argdefault.switch)
            kwargs[argname] = v

    if "args" in function_args:
        kwargs["args"] = args

    if "cq" in function_args:
        kwargs["cq"] = cq

    if "non_empty_job_list" in function_args:
        if not args:
            msg = f"The command {command_name!r} requires a non empty list of jobs as argument."
            raise UserError(msg)

        job_list = parse_job_list(args, context=context, cq=cq)

        # TODO: check non empty
        job_list = list(job_list)
        CompmakeConstants.aliases["last"] = job_list
        kwargs["non_empty_job_list"] = job_list

    if "job_list" in function_args:
        job_list = parse_job_list(args, context=context, cq=cq)
        job_list = list(job_list)
        CompmakeConstants.aliases["last"] = job_list
        # TODO: this does not survive reboots
        # logger.info('setting alias "last"' )
        kwargs["job_list"] = job_list

    if "context" in function_args:
        kwargs["context"] = context

    for x in args_without_default:
        if not x in kwargs:
            msg = "Required argument %r not given." % x
            raise UserError(msg)

    try:
        res = function(**kwargs)
        if (res is not None) and (res != 0):
            msg = "Command %r failed: %s" % (commands_line, res)
            raise CommandFailed(msg)
        return None
    finally:
        if dbchange:
            cq.invalidate()


def get_defaults(signature: inspect.Signature) -> Dict[str, object]:
    defaults = {}
    for k, v in signature.parameters.items():
        if v.default != inspect.Parameter.empty:
            defaults[k] = v.default

    return defaults


def get_args_without_defaults(signature: inspect.Signature) -> List[str]:
    args_without_default = []
    for k, v in signature.parameters.items():
        if v.default == inspect.Parameter.empty:
            args_without_default.append(k)
    return args_without_default



from typing import cast, Optional, Union

from compmake import Promise
from compmake.jobs import get_job_cache
from compmake.jobs.uptodate import CacheQueryDB
from compmake.structures import Cache, CMJobID

__all__ = [
    "compmake_execution_stats",
]


# @contract(promise="unicode|isinstance(Promise)", returns=Promise)
from zuper_commons.types import check_isinstance


def compmake_execution_stats(context, promise: Union[CMJobID, Promise], use_job_id: Optional[CMJobID] = None):
    """
        Returns a promise for a the execution stats of a job
        and its dependencies.
    """
    check_isinstance(promise, (Promise, str))

    if isinstance(promise, Promise):
        job_id = promise.job_id
    elif isinstance(promise, str):
        job_id = cast(CMJobID, promise)
        promise = Promise(job_id)
    else:
        assert False

    p2 = context.comp(dummy, promise)
    if use_job_id is not None:
        context.comp_prefix(None)
        return context.comp_dynamic(count_resources, the_job=job_id, job_id=use_job_id, extra_dep=p2)
    else:
        return context.comp_dynamic(count_resources, the_job=job_id, extra_dep=p2)


def dummy(*args):
    pass


def count_resources(context, the_job):
    db = context.get_compmake_db()
    cache = get_job_cache(the_job, db=db)
    if cache.state != Cache.DONE:
        msg = "The job %s was supposed to be finished: %s" % (the_job, cache)
        raise Exception(msg)

    cq = CacheQueryDB(db)
    children = cq.tree_children_and_uodeps(the_job)
    check_isinstance(children, set)
    children.add(the_job)

    res = {}
    for j in children:
        res[j] = context.comp_dynamic(
            my_get_job_cache, j, extra_dep=[Promise(j)], job_id="count-%s-%s" % (the_job, j)
        )

    return context.comp(finalize_result, res)


def my_get_job_cache(context, the_job):
    """ Gets the job cache, making sure it was done """
    db = context.get_compmake_db()
    cache = get_job_cache(the_job, db=db)
    if cache.state != Cache.DONE:
        msg = "The job %s was supposed to be finished: %s" % (the_job, cache)
        raise Exception(msg)
    return cache


# @contract(caches='list[>=1]')
def finalize_result(res):
    def stats_from_cache(cache: Cache):
        check_isinstance(cache, Cache)
        return dict(walltime=cache.walltime_used, cputime=cache.cputime_used)

    jobs = set(res.keys())
    allstats = list(map(stats_from_cache, res.values()))

    cpu_time = sum(x["cputime"] for x in allstats)
    wall_time = sum(x["walltime"] for x in allstats)
    return dict(cpu_time=cpu_time, wall_time=wall_time, jobs=jobs)



""" The actual interface of some commands in commands.py """

from compmake.plugins.console_output import write_line_endl
from zuper_commons.types import check_isinstance
from ..jobs import (
    children,
    direct_children,
    direct_parents,
    get_job,
    get_job_args,
    get_job_cache,
    job_args_sizeof,
    job_cache_exists,
    job_cache_sizeof,
    job_userobject_exists,
    job_userobject_sizeof,
    parents,
)
from ..structures import Cache
from ..ui import compmake_colored, ui_command, VISUALIZATION


@ui_command(section=VISUALIZATION, alias="lsl")
def details(non_empty_job_list, context, cq, max_lines=None):
    """ Shows the details for the given jobs including
        dependencies and stderr/stdout.

        The stderr/stdout lines are truncated. Use the ``max_lines``
        argument to see more:

            details max_lines=1000
    """
    num = 0
    for job_id in non_empty_job_list:
        # insert a separator if there is more than one job
        if num > 0:
            print("-" * 74)
        list_job_detail(job_id, context, cq, max_lines=max_lines)
        num += 1


def list_job_detail(job_id, context, cq, max_lines):
    db = context.get_compmake_db()

    dparents = direct_parents(job_id, db=db)
    all_parents = parents(job_id, db=db)
    other_parents = set(all_parents) - set(dparents)

    # TODO: use quicker up to date
    up, reason, _ = cq.up_to_date(job_id)

    red = lambda x: compmake_colored(x, "red")
    bold = lambda x: compmake_colored((x + " ").rjust(15), attrs=["bold"])

    def format_list(x):
        return "\n- ".join([""] + sorted(x))

    job = get_job(job_id, db=db)

    # TODO: make it work in Python3K
    print(bold("Job ID:") + "%s" % job_id)
    print(bold("Defined by:") + "%s" % job.defined_by)

    job_args = get_job_args(job_id, db=db)
    command, args, kwargs = job_args
    print(bold("command:") + "%s" % command)

    dchildren = direct_children(job_id, db=db)
    print(bold("Dependencies: (direct)") + " (%d) " % len(dchildren) + format_list(dchildren))

    all_children = children(job_id, db=db)
    other_children = set(all_children) - set(dchildren)
    print(bold("Dependencies: (other)") + " (%d) " % len(other_children) + format_list(other_children))

    print(bold("Dependencies: (dynamic)") + "%s" % job.dynamic_children)

    print(bold("Depending on this (direct):") + format_list(dparents))
    print(bold("Depending on this (other):") + format_list(other_parents))

    if job_cache_exists(job_id, db=db):
        cache2 = get_job_cache(job_id, db=db)

        print(bold("Status:") + "%s" % Cache.state2desc[cache2.state])
        print(bold("Uptodate:") + "%s (%s)" % (up, reason))

        if cache2.state == Cache.DONE:  # and cache.done_iterations > 1:
            # print(bold('Iterations:') + '%s' % cache.done_iterations)
            print(bold("Wall Time:") + "%.4f s" % cache2.walltime_used)
            print(bold("CPU Time:") + "%.4f s" % cache2.cputime_used)

            print("making: %s" % cache2.int_make)
            print("-- load: %s" % cache2.int_load_results)
            print("-- comp: %s" % cache2.int_compute)
            print("--   GC: %s" % cache2.int_gc)
            print("-- save: %s" % cache2.int_save_results)

            print(bold("Host:") + "%s" % cache2.host)

            if not job_userobject_exists(job_id, db):
                print(red("inconsistent DB: user object does not exist."))
    else:
        print(bold("Status:") + "%s" % Cache.state2desc[Cache.NOT_STARTED])
        cache2 = None

    jobargs_size = job_args_sizeof(job_id, db)
    print(bold("      args size: ") + "%s" % jobargs_size)

    if job_cache_exists(job_id, db):
        cache_size = job_cache_sizeof(job_id, db)
        print(bold("     cache size: ") + "%s" % cache_size)
    else:
        cache_size = 0

    if job_userobject_exists(job_id, db):
        userobject_size = job_userobject_sizeof(job_id, db)
        print(bold("userobject size: ") + "%s" % userobject_size)
    else:
        userobject_size = 0

    total = jobargs_size + cache_size + userobject_size
    print(bold("          Total: ") + "%s" % total)

    def display_with_prefix(buffer, prefix="", transform=lambda x: x):  # @ReservedAssignment
        check_isinstance(buffer, str)
        lines = buffer.split("\n")
        if max_lines is not None:
            if len(lines) > max_lines:
                warn = ".... Showing only last %d lines of %d ... " % (max_lines, len(lines))
                lines = [warn] + lines[-max_lines:]

        for line in lines:
            s = "%s%s" % (prefix, transform(line))
            write_line_endl(s)
            # if six.PY2:
            # s = s.encode('utf-8')
            # sys.stdout.buffer.write(s)

    if cache2 is not None:
        stdout = cache2.captured_stdout
        if stdout and stdout.strip():
            display_with_prefix("-----> captured stdout <-----")
            display_with_prefix(stdout, prefix="|")

        stderr = cache2.captured_stderr
        if stderr and stderr.strip():
            display_with_prefix("-----> captured stderr <-----")
            display_with_prefix(stderr, prefix="|")

        if cache2.state == Cache.FAILED:
            display_with_prefix(cache2.exception, prefix="exc |")
            display_with_prefix(cache2.backtrace, prefix="btr |")



import sys

from ..events import register_handler
from ..state import get_compmake_config
from ..ui import compmake_colored
from ..utils import get_screen_columns


stream = sys.stderr

counter = 0


def console_write(s):
    """ Writes a line that will be erased. """
    cols = get_screen_columns()
    s = s.ljust(cols)
    stream.write(s)
    stream.write("\r")


def job_redefined(context, event):
    if not get_compmake_config("verbose_definition"):
        return
    stream.write(compmake_colored("Redefined %s\r" % event.job_id, "yellow", attrs=["bold"]))
    stream.write(compmake_colored(event.reason, "yellow"))
    # stream.write('\n')


def job_defined(context, event):
    if not get_compmake_config("verbose_definition"):
        return
    global counter
    counter += 1
    console_write("compmake: defining job #%d %s" % (counter, event.job_id))


register_handler("job-redefined", job_redefined)
register_handler("job-defined", job_defined)

# register_handler('job-already-defined', lambda event:
# console_write('Confirming job %s' % event.job_id))



from typing import Sequence

from zuper_commons.ui import get_colorize_function

__all__ = [
    "compmake_colored",
]


class ColoredCached:
    functions = {}


def compmake_colored(x: str, color: str = None, on_color: str = None, attrs: Sequence[str] = None) -> str:
    from .. import get_compmake_config

    colorize = get_compmake_config("colorize")
    if not colorize:
        return x

    if color is None and on_color is None:
        return x

    key = (color, on_color, tuple(attrs or ()))

    if key not in ColoredCached.functions:
        ColoredCached.functions[key] = get_colorize_function(rgb=color, bg_color=on_color)

    f = ColoredCached.functions[key]

    colorize = get_compmake_config("colorize")
    if colorize:
        return f(x)
        # return termcolor_colored(x, color, on_color, attrs)
    else:
        return x



import sys

from six import StringIO

# from contracts import check_isinstance
from .coloredterm import termcolor_colored
from .strings_with_escapes import pad_to_screen

RESET = "\033[0m"  # XXX

__all__ = [
    "OutputCapture",
]


class LineSplitter:
    """ A simple utility to split an incoming sequence of chars
        in lines. Push characters using append_chars() and
        get the completed lines using lines(). """

    def __init__(self):
        self.current = ""
        self.current_lines = []

    def append_chars(self, s: str):
        assert isinstance(s, str), s
        for char in s:
            if char == "\n":
                self.current_lines.append(self.current)
                self.current = ""
            else:
                self.current += char

    def lines(self):
        """ Returns a list of line; empties the buffer """
        l = list(self.current_lines)
        self.current_lines = []
        return l


class StreamCapture(object):
    def __init__(self, transform=None, dest=None, after_lines=None):
        """ dest has write() and flush() """
        self.buffer = StringIO()
        self.dest = dest
        self.transform = transform
        self.line_splitter = LineSplitter()
        self.after_lines = after_lines

    def write(self, s: str):

        assert isinstance(s, str), s
        self.buffer.write(s)
        self.line_splitter.append_chars(s)
        lines = self.line_splitter.lines()

        if self.dest:
            # XXX: this has a problem with colorized things over multiple lines
            for line in lines:
                if self.transform:
                    line = self.transform(line)
                self.dest.write("%s\n" % line)

            # self.dest.write(self.transform(s))
            self.dest.flush()

        if self.after_lines is not None:
            self.after_lines(lines)

    def get_value_text_type(self):
        b = self.buffer.getvalue()

        return b

    def flush(self):
        pass


# TODO: this thing does not work with logging enabled
class OutputCapture(object):
    def __init__(self, context, prefix, echo_stdout=True, echo_stderr=True):
        self.old_stdout = sys.stdout
        self.old_stderr = sys.stderr

        from ..events import publish

        def publish_stdout(lines):
            publish(context, "job-stdout", job_id=prefix, lines=lines)

        def publish_stderr(lines):
            publish(context, "job-stderr", job_id=prefix, lines=lines)

        # t1 = lambda s: '%s|%s' % (prefix, colored(s, 'cyan', attrs=['dark']))

        # FIXME: perhaps we should use compmake_colored
        t1 = lambda s: "%s|%s" % (termcolor_colored(prefix, attrs=["dark"]), s)
        t2 = lambda s: RESET + pad_to_screen(t1(s))
        dest = {True: sys.stdout, False: None}[echo_stdout]
        self.stdout_replacement = StreamCapture(transform=t2, dest=dest, after_lines=publish_stdout)
        sys.stdout = self.stdout_replacement

        # t3 = lambda s: '%s|%s' % (prefix, colored(s, 'red', attrs=['dark']))
        t3 = lambda s: "%s|%s" % (termcolor_colored(prefix, "red", attrs=["dark"]), s)
        t4 = lambda s: RESET + pad_to_screen(t3(s))
        dest = {True: sys.stderr, False: None}[echo_stderr]
        self.stderr_replacement = StreamCapture(transform=t4, dest=dest, after_lines=publish_stderr)
        sys.stderr = self.stderr_replacement

    def deactivate(self):
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr

    def get_logged_stdout(self):
        return self.stdout_replacement.get_value_text_type()  # buffer.getvalue()

    def get_logged_stderr(self):
        return self.stderr_replacement.get_value_text_type()  # buffer.getvalue()



import pickle as compmake_pickle

__all__ = [
    "try_pickling",
]


def try_pickling(obj):
    """ Serializes and deserializes an object. """
    s = compmake_pickle.dumps(obj)
    compmake_pickle.loads(s)



import time
from typing import Tuple, Union

from zuper_commons.types import describe_type
from ..structures import ProgressStage


class Globals:
    stack = []
    callbacks = []


def progress_stack_updated():
    for callback in Globals.callbacks:
        callback(Globals.stack)


def init_progress_tracking(my_callback):
    Globals.stack = []
    Globals.callbacks = []  # OK, otherwise old callbacks will update status
    Globals.callbacks.append(my_callback)
    progress_stack_updated()


# @contract(taskname="unicode", iterations="tuple(int|float,int|float)")
def progress(taskname: str, iterations: Tuple[Union[int, float], Union[int, float]], iteration_desc=None):
    """
        Function used by the user to describe the state of the computation.

       Parameters
       ---------

       - ``name``: must be a string, describing the current task.

       - ``iterations``: must be a tuple of two integers (k,N), meaning
          that the current iteration is the k-th out of N.

       - ``iteration_desc``: an optional string describing the current iteration



            for i in range(n):
                progress('Reading files', (i,n),
                         'processing file %s' % file[i])
    """

    if not isinstance(taskname, str):
        raise ValueError(
            "The first argument to progress() is the task name "
            + "and must be a string; you passed a %s." % describe_type(taskname)
        )

    if not isinstance(iterations, tuple):
        raise ValueError(
            "The second argument to progress() must be a tuple,"
            + " you passed a %s." % describe_type(iterations)
        )
    if not len(iterations) == 2:
        raise ValueError(
            "The second argument to progress() must be a tuple "
            + " of length 2, not of length %s." % len(iterations)
        )

    if not isinstance(iterations[0], (int, float)):
        raise ValueError(
            "The first element of the tuple passed to progress "
            + "must be integer or float, not a %s." % describe_type(iterations[0])
        )

    if not iterations[1] is None and not isinstance(iterations[1], (int, float)):
        raise ValueError(
            "The second element of the tuple passed to progress "
            "must be either None or an integer, not a %s." % describe_type(iterations[1])
        )

    if iterations[1] < iterations[0]:
        raise ValueError("Invalid iteration tuple: %s" % str(iterations))

    BROADCAST_INTERVAL = 0.5

    is_last = iterations[0] == iterations[1] - 1

    stack = Globals.stack

    for i, stage in enumerate(stack):
        if stage.name == taskname:
            # remove children
            has_children = i < len(stack) - 1
            stack[i + 1 :] = []
            stage.iterations = iterations
            stage.iteration_desc = iteration_desc
            # TODO: only send every once in a while
            if (
                (is_last or has_children)
                or (stage.last_broadcast is None)
                or (time.time() - stage.last_broadcast > BROADCAST_INTERVAL)
            ):
                progress_stack_updated()
                stage.last_broadcast = time.time()
            if stage.last_broadcast is None:
                stage.last_broadcast = time.time()
            break
    else:
        # If we are here, we haven't found taskname in the stack.
        # This means that it is either a child or a brother (next task)
        # We check that the last stage was over
        while stack and stack[-1].was_finished():
            stack.pop()
            # treat it as a brother

        stack.append(ProgressStage(taskname, iterations, iteration_desc))
        progress_stack_updated()



# -*- coding: utf-8 -*-

from compmake.state import get_compmake_config

__all__ = ["disable_logging_if_config"]


def disable_logging_if_config(context):
    """ Disables Multyvac's logging if specified in config. """

    import logging

    if not get_compmake_config("multyvac_debug"):
        logging.getLogger("multyvac").setLevel(logging.WARNING)
